#include <QDir>

#include "application.h"

Application * Application::fileManager = nullptr;
QStringList Application::validApplications = QStringList();
QStringList Application::groupCache = QStringList();
QStringList Application::groupIdCache = QStringList();
QList< Application* > Application::appList = QList< Application * >();
QHash< QString,QList<Application*> > Application::appCache = QHash< QString,QList<Application*> >();
QHash< QString,QList<Application*> > Application::appByGroupCache = QHash< QString,QList<Application*> >();
QHash< QString,QList<Application*> > Application::groupByAppCache = QHash< QString,QList<Application*> >();
QHash< QString,QByteArray > Application::groupIconCache = QHash< QString,QByteArray >();
QHash< QString,QStringList > Application::extensionsCache = QHash< QString,QStringList >();
QString Application::socketDir = QString();

Application::Application( const QString & id )
{
	pid = 0;
	this->id = id;
	fileSystemWatcher.addPath( Settings::homePath() + QDir::separator() + ".liberus" + QDir::separator() + "run");
}

Application::Application( Application * app )
{
	id = app->getId();
	name = app->getName();
	description = app->getDescription();
	command = app->getCommand();
	server = app->getServer();
	serverId = app->getServerId();
	os = app->getOS();
	icon = app->getIcon();
	arguments = app->getArguments();
	workingPath = app->getWorkingPath();
	file = app->getFile();
	groups = app->getGroups();
	mimeTypes = app->getMimeTypes();

	fileSystemWatcher.addPath( Settings::homePath() + QDir::separator() + ".liberus" + QDir::separator() + "run");

	connect( this, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(deleteAppInstance()) );
}

void Application::deleteApplicationCache()
{
	deleteList( appList );
	groupCache.clear();
	groupIdCache.clear();
	groupIconCache.clear();
	appCache.clear();
	appByGroupCache.clear();
	groupByAppCache.clear();
	validApplications.clear();
	extensionsCache.clear();
}

Application * Application::getFileManager()
{
	if ( fileManager == nullptr )
	{
		setFileManager();
	}

	return fileManager;
}
void Application::setFileManager()
{
	fileManager = new Application(tr("Liberus Files"));
#ifdef Q_OS_LINUX
	QString fmCommand = Settings::getValue("Advanced/PreferredFileManager");
	if ( fmCommand.isEmpty() )
	{
		fmCommand = "liberus-desktop";
		fileManager->setArguments( QStringList("-f") );
	}
	fileManager->setCommand( fmCommand );
#endif
#ifdef Q_OS_WIN
	fileManager->setCommand( "%SYSTEMROOT%\\explorer.exe");
	//fileManager->setCommand( SHAREDIR + "liberus-desktop.exe" );
	//fileManager->setArguments( QStringList("-f") );
#endif

}
void Application::updateApplicationCache(InitProgress * initProgress)
{
	deleteApplicationCache();

	if ( LWClient::instance()->isConnected() )
	{
		setFileManager();
		// Get groupIdCache
		QString res = LWClient::instance()->sendMessage("get_application_group_ids" );
		if (! res.startsWith("Error:") )
		{
			groupIdCache = res.split(":");
		}
		qDebug("groupIdCache: %s", groupIdCache.join(",").toUtf8().data());

		// Get groupCache
		res = LWClient::instance()->sendMessage("get_application_groups" );
		if (! res.startsWith("Error:") )
		{
			QStringList groupList = res.split(":");
			groupCache = groupList.replaceInStrings(QRegExp("^\\s*"),"");
		}
		qDebug("groupCache: %s", groupCache.join(",").toUtf8().data());

		QString appListString = LWClient::instance()->sendMessage("get_valid_applications");
		if (! appListString.startsWith("Error:") && ! appListString.simplified().isEmpty())
		{
			validApplications = appListString.split(",");
			Settings::logInfo( "Valid applications: " + validApplications.join(",") );
			for ( int i = 0; i < validApplications.count(); i++ )
			{
				if ( initProgress != nullptr )
				{
					initProgress->setProgress( (100 * i) / validApplications.count() );
				}
				Application * app = new Application( validApplications[i] );
				setServerValues( validApplications[i], app );
				QString extensionsString = LWClient::instance()->sendMessage("get_application_extensions?" + app->getId());
				if ( ! res.startsWith("Error:") )
				{
					QStringList extensions = extensionsString.split(":");
					for ( int i = 0; i < extensions.count(); i++ )
					{
						appCache[extensions[i]] += app;
					}
				}
				appList += app;
				if ( ! appByGroupCache.contains("all") )
				{
					appByGroupCache["all"] = QList<Application*>();
				}
				appByGroupCache["all"] += app;
				QString applicationGroupsString = LWClient::instance()->sendMessage("get_application_group_ids?" + app->getId());
				if ( ! res.startsWith("Error:") )
				{
					QStringList appGroupList = applicationGroupsString.split(",");
					if ( appGroupList.count() == 0 )
					{
						appGroupList += "other";
					} else {
						if ( appGroupList.count() == 1 && appGroupList[0].isEmpty() )
						{
							appGroupList[0] = "other";
						}
					}
					for ( int k = 0; k < appGroupList.count(); k++ )
					{
						if ( ! appGroupList[k].isEmpty() )
						{
							appByGroupCache[appGroupList[k]] += app;
						}
					}
				}
			}
		}
	} else {
#ifdef Q_OS_LINUX
		groupIdCache = Settings::getIds(GROUPS_FILE);
		groupCache = Settings::getValues( GROUPS_FILE, "Name" );
		QFile mimeFile( "/etc/mime.types" );
		if (mimeFile.open(QIODevice::ReadOnly))
		{
			QTextStream in(&mimeFile);
			while (!in.atEnd())
			{
				QString line = in.readLine().simplified();
				if ( ( ! line.startsWith("#") ) && ( ! line.isEmpty() ) )
				{
					QString mimeType = line.section(" ", 0, 0);
					QStringList extensions = line.split(" ");
					extensions.removeAll( mimeType );
					if ( extensions.count() > 0 )
					{
						extensionsCache[mimeType] = extensions;
					}
				}
			}
			mimeFile.close();
		}

		QStringList appNameList;
		QString apps = BASEDIR "/share/applications";
		QDir appsDir( apps );
		QStringList appsDirEntries = appsDir.entryList(QStringList(), QDir::AllDirs);

		for ( int i = 0; i < appsDirEntries.count(); i++ )
		{
			if ( initProgress != 0 )
			{
				initProgress->setProgress( (100 * i) / appsDirEntries.count() );
			}
			if ( appsDirEntries[i] != ".." )
			{
				QDir appsDirDir( apps + QDir::separator() + appsDirEntries[i] );
				QStringList filters;
				filters << "*.desktop";
				QStringList files = appsDirDir.entryList( filters );
				for ( int i = 0; i < files.count(); i++)
				{
					Application * app = fromDesktopFile(appsDirDir.absolutePath() + QDir::separator() + files[i], appNameList);
					if ( app != 0 )
					{
						validApplications += app->getName();
						appNameList += app->getName();
						appList += app;
						/*if ( app->getName() == "Liberus Files" )
						{
							fileManager = app;
						}*/
						QStringList mimeList = app->getMimeTypes();
						QStringList extensions;
						for ( int i = 0; i < mimeList.count(); i++ )
						{
							extensions += extensionsCache[mimeList[i]];
						}
						for ( int i = 0; i < extensions.count(); i++ )
						{
							appCache[extensions[i]] += app;
						}

						QStringList appGroups;
						for ( int i = 0; i < app->getGroups().count(); i++ )
						{
							if ( groupIdCache.contains( app->getGroups()[i] ) )
							{
								appGroups += app->getGroups()[i];
							}
						}
						if ( appGroups.count() == 0 )
						{
							appGroups += "other";
						}
						appGroups += "all";

						for ( int i = 0; i < appGroups.count(); i++ )
						{
							if ( appByGroupCache.contains( appGroups[i] ) )
							{
								appByGroupCache[appGroups[i]] += app;
							} else {
								appByGroupCache[appGroups[i]] = QList<Application*>();
								appByGroupCache[appGroups[i]] += app;
							}
						}
					}
				}
			}
		}
#endif
#ifdef Q_OS_WIN
		QStringList groupIds = applicationGroupIdsList();
		QStringList validApps = Settings::getApplicationsList();
		Settings::logInfo( "Valid applications: " + validApps.join(",") );
		for ( int i = 0; i < validApps.count(); i++ )
		{
			if ( initProgress != nullptr )
			{
				initProgress->setProgress( (100 * i) / validApps.count() );
			}
			Application * app = new Application( validApps[i] );
			setLocalValues( validApps[i], app );
			//appCache[extensions[i]] += app;
			QStringList appGroups = applicationGroupIdsList();
			qDebug("Application: %s, app groups: %s", validApps[i].toUtf8().data(), appGroups.join(" ").toUtf8().data());
			for ( int i = 0; i < appGroups.count(); i++ )
			{
				if ( appByGroupCache.contains( appGroups[i] ) )
				{
					appByGroupCache[appGroups[i]] += app;
				} else {
					appByGroupCache[appGroups[i]] = QList<Application*>();
					appByGroupCache[appGroups[i]] += app;
				}
			}
		}
#endif
	}
}

Application * Application::fromDesktopFile(QString desktopFile, QStringList & appNameList)
{
	QMap<QString,QString> settings = Settings::desktopSettings(desktopFile);
	QString name = settings.value("Desktop Entry/Name");
	if ( appNameList.contains( name ) )
	{
		return nullptr;
	}
	if ( settings.value("Desktop Entry/NoDisplay").compare("True", Qt::CaseInsensitive) == 0 )
	{
		return nullptr;
	}
	if ( settings.value("Desktop Entry/Hidden").compare("True", Qt::CaseInsensitive) == 0 )
	{
		return nullptr;
	}
	if ( ! settings.value("Desktop Entry/OnlyShowIn").isEmpty() )
	{
		return nullptr;
	}
	if ( settings.value("Desktop Entry/Icon").isEmpty() )
	{
		return nullptr;
	}
	QStringList groups = settings.value("Desktop Entry/Categories").split(";");
	groups.removeAll("");
	if ( groups.count() == 0 )
	{
		return nullptr;
	}
	appNameList += name;
	Application * app = new Application( name );
	QString description = settings.value("Desktop Entry/Comment");

	if ( description.isEmpty() )
	{
		description = name;
	}
	app->setName( name );
	app->setDescription( description);
	app->setGroups( groups );

	QString command = settings.value("Desktop Entry/Exec");
	command = command.replace("%U", "");
	command = command.replace("%u", "");
	command = command.replace("%F", "");
	command = command.replace("%f", "");
	command = command.replace("%i", "");
	command = command.replace("%c", name);
	command = command.simplified();
	QStringList arguments = command.split(" ");
	if ( arguments.count() > 1 )
	{
		command = arguments[0];
		arguments.removeAt(0);
		app->setArguments( arguments );
	}
	app->setCommand( command );
	app->setMimeTypes( settings.value("Desktop Entry/MimeType").split(";") );
	app->setIcon( Application::findIcon( settings.value("Desktop Entry/Icon") ) );
	app->setServer("localhost");
	app->setOS(RUNNING_OS);
	app->setProtocol(QString());

	return app;
}

QIcon Application::findIcon(QString iconName)
{
	QIcon icon;
	QStringList iconPath;
	QString iconFile;

#ifdef Q_OS_LINUX
	QStringList sharePaths;
	sharePaths << BASEDIR "/share/icons/default";
	sharePaths << BASEDIR "/share/icons/hicolor";
	sharePaths << BASEDIR "/share/icons/gnome";
	sharePaths << BASEDIR "/share/icons/oxygen/base";
	sharePaths << BASEDIR "/share/icons/Adwaita";

	QStringList sizes;
	sizes << "128x128";
	sizes << "64x64";
	sizes << "48x48";
	sizes << "32x32";
	sizes << "22x22";

	QStringList categories;
	categories << "apps";
	categories << "actions";
	categories << "categories";
	categories << "status";
	categories << "places";
	categories << "mimetypes";
	categories << "devices";
	categories << "emblems";

	QStringList formats;
	formats << ".png";
	formats << ".xpm";
	formats << ".svg";

	iconPath += "";
	iconPath += BASEDIR "/share/pixmaps";

	for ( int format = 0; format < formats.count(); format++ )
	{
		for ( int category = 0; category < categories.count(); category++ )
		{
			for ( int size = 0; size < sizes.count(); size++ )
			{
				for ( int path = 0; path < sharePaths.count(); path++ )
				{
					iconPath += sharePaths[path] + "/" + sizes[size] + "/" + categories[category];
				}
			}
		}
	}

	bool found = false;
	for ( int format = 0; format < formats.count() && ! found; format++ )
	{
		for ( int i = 0; i < iconPath.count() && ! found; i++ )
		{
			iconFile = iconPath[i] + "/" + iconName;
			if ( ! iconFile.endsWith(formats[format]) )
			{
				iconFile += formats[format];
			}
			if ( QFile::exists( iconFile ) )
			{
				found = true;
			} else {
				iconFile = QString();
			}
		}
	}
#endif
	if ( ! iconFile.isEmpty() )
	{
		return ( QIcon( iconFile ) );
	} else {
		return QIcon();
	}
}

QString Application::windowsPath(QString file)
{
	QString res;

	res = file.replace(SHARED_DIR, Settings::windowsSharedPath());
	res = res.replace( Settings::homePath(),  Settings::windowsHomePath());
	res = res.replace('/', '\\');

	return res;
}

QString Application::mimeType(QFileInfo fileInfo)
{
	int count = 0;
	while (fileInfo.isSymLink() && count < MAXLINKDEPTH)
	{
		fileInfo = QFileInfo(fileInfo.symLinkTarget());
		count++;
	}

	if (fileInfo.isSymLink())
		return QString("recursivelink");

	QString ext = fileInfo.suffix();
	QFile extensions (MIME_DIR + QString(QDir::separator()) + "extensions");
	//if ( ! extensions.open(QIODevice::ReadOnly | QIODevice::Text) )
	if ( ! extensions.open(QIODevice::ReadOnly) )
		return QString("application/octet-stream");

	while ( ! extensions.atEnd() )
	{
		QByteArray line = extensions.readLine();
 
		if ( line.startsWith( ext.toUtf8().data() ) )
		{
			line.replace( QString( ext + ": " ).toUtf8().data(), "" );
			return QString( line ).trimmed();
		}
	}

	return QString("application/octet-stream");

/* With magic mime:

	QMimeDatabase mime;

	int count = 0;
	while (fileInfo.isSymLink() && count < MAXLINKDEPTH)
	{
		fileInfo = QFileInfo(fileInfo.symLinkTarget());
		count++;
	}

	if (fileInfo.isSymLink())
		return QString("recursivelink");

	QString res = mime.mimeTypeForFile(fileInfo).name();

	return res.trimmed();
*/
}

QIcon Application::getIcon(QFileInfo file)
{
	QString extension = file.suffix();
	if ( appCache.contains( extension ) )
	{
		if ( appCache[extension].count() > 0 )
		{
			return appCache[extension][0]->getIcon();
		}
	}
	return QIcon();
}

Application * Application::preferredProgram(QString extension)
{
	QList<Application*> apps = appCache[extension];
	if ( apps.count() > 0 )
	{
		return apps[0];
	} else {
		return nullptr;
	}
}

QStringList Application::applicationGroupsList()
{
	return groupCache;
}

QStringList Application::applicationGroupIdsList()
{
	return groupIdCache;
}

QIcon Application::applicationGroupIcon(QString groupId)
{
	QByteArray ic;

	if ( LWClient::instance()->isConnected() )
	{
		if ( groupIconCache.contains( groupId ) )
		{
			ic = groupIconCache[groupId];
		} else {
			ic = LWClient::instance()->sendMessageByteArray( "get_application_group_icon?" + groupId );
			groupIconCache[groupId] = ic;
		}
		QPixmap pixmap;
		pixmap.loadFromData( ic );
		return QIcon( pixmap );
	} else {
		QString icon = Settings::getValue(GROUPS_FILE, groupId + "/Icon");
		return QIcon(":/icons/groups/" + icon + ".png");
	}
}

QList<Application*> Application::applicationListByGroup(QString groupId)
{
	if (appByGroupCache.contains(groupId) )
	{
		return appByGroupCache[ groupId ];
	}

	QList<Application*> res;

	QStringList validApplications = getValidApplications();
	QStringList appListByGroup;


	if ( LWClient::instance()->isConnected() )
	{
		QString resString = LWClient::instance()->sendMessage("get_applications_by_group?" + groupId );
        if (! resString.startsWith("Error:") )
            appListByGroup = resString.split(":");
        int count = appListByGroup.count();
        for (int i = 0; i < count; i++)
        {
            if ( validApplications.contains( appListByGroup[i] ) )
            {
                res.append(new Application(appListByGroup[i]));
            }
        }
    } else {
#ifdef Q_OS_WIN
        for (int i = 0; i < validApplications.count(); i++)
        {
            QStringList groups = Settings::getValue(APPLICATIONS_FILE, validApplications[i] + "/groups").split(":");
            if (groups.contains( groupId))
            {
                res.append(new Application(validApplications[i]));
            }
        }
#endif
    }
	appByGroupCache[ groupId ] = res;
	return res;
}

QList<Application*> Application::applicationDeletedList(QString extension)
{
	QList<Application*> list;

    QStringList homeAppList = Settings::getValue( MIME_DIR + QString(QDir::separator()) + "extensions", extension).split(":");
	int count = homeAppList.count();
	for (int i = 0; i < count; i++)
	{
		if (homeAppList[i].startsWith("!"))
		{
			list += new Application(homeAppList[i].right( homeAppList[i].length() - 1));
		}
	}

	return list;
}

QList<Application*> Application::favoriteApplicationList()
{
	bool isConnected = LWClient::instance()->isConnected();
	QList<Application*> apps;
	QStringList favoritesStringList = Settings::getFavoriteApplications();
	QList<Application*> all = appList;
	for ( int i = 0; i < all.count(); i++ )
	{
		if ( isConnected || favoritesStringList.contains(all[i]->getName()))
		{
			apps += all[i];
		}
	}
	return apps;
}

QList<Application*> Application::applicationList(QString extension)
{
	if ( extension.isNull() ) // return all applications
	{
		return appList;
	}

	if ( appCache.contains( extension ) )
	{
		return appCache[extension];
	}

	QList<Application*> list;
	QStringList apps = getValidApplications();

	//QStringList homeAppList = Settings::getValue( EXTENSIONS_FILE, extension).split(":");

	QStringList appStringList;

	if ( LWClient::instance()->isConnected() )
	{
		QString res = LWClient::instance()->sendMessage("get_applications?" + extension );
		if (! res.startsWith("Error:") )
		{
			appStringList = res.split(":");
			for ( int i = 0; i < appStringList.count(); i++ )
			{
				list += new Application( appStringList[i] );
			}
		}
	} else {
#ifdef Q_OS_WIN
		QString type = Settings::getMimetype( extension );
		QStringList applications = Settings::getApplications( type.simplified() );
		for (int i = 0; i < applications.count(); i++)
		{
			if (apps.contains( applications[i] ))
			{
				list += new Application( applications[i] );
			}
		}
#endif
	}


	/*
	QStringList delList;

	int count = homeAppList.count();
	for (int i = 0; i < count; i++)
	{
		if (homeAppList[i].startsWith("!"))
		{
			QString deletedApp = homeAppList[i].right( homeAppList[i].length() - 1);
			appStringList.removeAll(deletedApp);
			homeAppList.removeAll(homeAppList[i]);
			count = homeAppList.count();
			i--;
		}
	}

	homeAppList += appStringList;

	homeAppList.removeDuplicates();

	count = homeAppList.count();
	for (int i = 0; i < count; i++)
	{
		if (apps.contains(homeAppList[i]))
		{
			list.append(new Application(homeAppList[i]));
		}
	}
	*/

	appCache[extension] = list;
	return list;
}

void Application::deleteList( QList<Application*> & list )
{
	for (int i = 0; i < list.count(); i++)
	{
		delete list[i];
	}
	list.clear();
}

QString Application::resolveWindowsEnvironment( QString command )
{
    QString res = command.replace("\\", "\\\\");
    QStringList variables;
    variables += "%programw6432%";
    variables += "%programfiles(x86)%";
    variables += "%programfiles%";
    variables += "%windir%";
    variables += "%systemroot%";
    variables += "%system32%";
    variables += "%system%";
    variables += "%homepath%";
    variables += "%temp%";
    variables += "%tmp%";

    for ( int i = 0; i < variables.count(); i++ )
    {
        res = res.replace( variables[i], QString( qgetenv( variables[i].replace("%", "").toUtf8().data() )).replace("\\", "\\\\"), Qt::CaseInsensitive);
    }
    res = res.replace("\\\\", "\\");
    return res.replace("%", "");
}

bool Application::isRemote(QString server, int os)
{
	bool res = true;
	if ( Settings::isLocalhost( server ) && os == Settings::getLocalOs() )
	{
		res = false;
	}
	return res;
}

bool Application::isValid(QString command, QString server, int os, QString protocol)
{
    bool res = false;
    if ( QFile::exists( command ) )
    {
        if ( Settings::isLocalhost( server ) )
        {
			if ( os == Settings::getLocalOs() )
            {
                res = true;
            }
        } else {
            if ( ! protocol.isEmpty() )
            {
                res = true;
            }
        }
    } else {
        if ( ! Settings::isLocalhost( server ) && ! protocol.isEmpty() )
        {
            res = true;
        }
    }
    return res;
}

QStringList Application::getValidApplications()
{
	return Application::validApplications;
}

void Application::addExtension(QString extension)
{
    Settings::setValue( MIME_DIR + QString(QDir::separator()) + "extensions", extension, QString());
}

QStringList Application::getExtensionsList()
{
    QStringList res;

	if ( LWClient::instance()->isConnected() )
	{
		QString lwRes = LWClient::instance()->sendMessage("get_extensions_list");
		if ( ! lwRes.startsWith("Error:") )
			res += lwRes.split(",");
    } else {
        QFile extensionsFile(EXTENSIONS_FILE);
        if (extensionsFile.open(QIODevice::ReadOnly))
        {
           QTextStream in(&extensionsFile);
           while (!in.atEnd())
           {
               res += in.readLine();
           }
           extensionsFile.close();
        }
    }

	return res;
}

/*QString Application::getValue(QString key)
{
	QString primaryFile = Settings::homePath() + "/.liberus-desktop/mime/applications";
	QString secondaryFile = MIME_DIR + QString(QDir::separator()) + "applications";

	return Settings::getValue(primaryFile, secondaryFile, key);
}*/

QStringList Application::getArguments(QString command, QStringList arguments, QString file, int os, QString serverAddress )
{
	qDebug("Server address: %s, os: %d", serverAddress.toUtf8().data(), os);
	if ( ( ! serverAddress.isEmpty() ) && ( os > 0 ) && isRemote(serverAddress, os) )
	{
		return remoteApplicationArguments( command, arguments, file, serverAddress );
	} else {
		QStringList args;
		if ( Settings::getBooleanValue( "ChrootEnvironment" ) )
		{
			args += "-p"; // Preserve environment on schroot
			args += "-c";
			args += "chroot:" + Settings::getUserName();
			args += "--"; // No more options
			args += "liberuslauncher.sh";
			args += command.trimmed().split(' ');
			args += arguments;
		} else {
#ifdef Q_OS_WIN
			args += resolveWindowsEnvironment( command );
#else
			args += command;
#endif
			args += arguments;
		}
		if ( ! file.isEmpty() )
		{
#ifdef Q_OS_WIN
			args += file.replace("/", "\\");
#endif
#ifdef Q_OS_LINUX
			args += file;
#endif
		}
		for ( int i = 0; i < args.count(); i++)
		{
			if ( args[i].isEmpty() )
			{
				args.removeAt(i);
				i--;
			}
		}
		return args;
	}
}

QStringList Application::remoteApplicationArguments(QString command, QStringList args, QString file, QString serverAddress )
{
	QStringList arguments;

//	QString shell = "%programfiles%\\Liberus Desktop\\fullscreenshell.exe";
//	QString shell = "c:\\fullscreenshell\\fullscreenshell.exe";
//	QString shell = "c:\\seamlessrdp\\seamlessrdpshell.exe";

	QString server = serverAddress;
	QHostAddress address = Settings::getLocalAddress();

	QString commandString = command.replace("\\\\", "\\") + args.join(" ").replace("\\\\", "\\");

	if ( QString( RDP_CLIENT ) == "xfreerdp" )
	{
		QString wm = Settings::getValue( "Advanced/PreferredWindowManager" );
		QStringList freeRDPOptions = Settings::getValue("Advanced/RemoteAppFreeRDPOptions").split(",");
		Settings::logInfo( QString("FreeRDP options: %1").arg( freeRDPOptions.join(",") ) );
		if ( freeRDPOptions.count() == 0 )
		{
			freeRDPOptions += "/sound:sys:alsa";
			//freeRDPOptions += "/gfx:rfx";
			freeRDPOptions += "/printer";
		}

		arguments += "/u:" + Settings::getUserName();
		QString domain = Settings::getValue("SambaDomain");
		if ( ! domain.isEmpty() )
		{
			arguments += "/d:" + domain;
		}
		arguments += "/p:" + Settings::getPassword();
		arguments += "/app:" + commandString;
		if ( ! file.isEmpty() )
		{
			arguments += "/app-file:" + Application::windowsPath(file).replace("\\\\", "\\");
		}
		arguments += freeRDPOptions;
		arguments += "/cert-ignore";
		arguments += "/v:" + server;
	}

	if ( QString( RDP_CLIENT ) == "rdesktop" )
	{
		QString keyboardLayout;
#ifdef Q_OS_LINUX
		QProcess * keyboardProcess = new QProcess();
		keyboardProcess->start( "setxkbmap", QStringList("-query") );
		if ( keyboardProcess->waitForStarted() )
		{
			keyboardProcess->waitForFinished();
			QList<QByteArray> output = keyboardProcess->readAll().split('\n');
			for ( int i = 0; i < output.count(); i++ )
			{
				if ( output[i].startsWith("layout:") )
				{
					keyboardLayout = QString( output[i] ).section(QRegExp("\\s+"), 1, 1);
				}
			}
		}
		delete keyboardProcess;
#endif
		QStringList rdesktopOptions = Settings::getValue("Advanced/RemoteAppRDesktopOptions").split(",");
		if ( rdesktopOptions.count() == 0 )
		{
			rdesktopOptions += "-r sound:local:alsa";
			rdesktopOptions += "-r clipboard:PRIMARYCLIPBOARD";
		}

		QString seamlessShell = Settings::getValue("Advanced/RemoteAppSeamlessRDPExecutable");
		if ( seamlessShell.simplified().isEmpty() )
		{
			seamlessShell = "%ProgramFiles%\\Liberus Server\\seamlessrdpshell.exe";
		}
		arguments << "-u" << Settings::getUserName();
		QString domain = Settings::getValue("Server/SambaDomain");
		if ( ! domain.isEmpty() )
		{
			arguments << "-d" << domain;
		}
		arguments << "-p" << Settings::getPassword();
		arguments << "-A" << seamlessShell;
		if ( socketDir.isEmpty() )
		{
			QTemporaryDir socketDirTemp;
			if ( socketDirTemp.isValid() )
			{
				socketDir = socketDirTemp.path();
			}
		}

		if ( ! socketDir.isEmpty() )
		{
			arguments << "-F" << socketDir;
		}

		if ( ! file.isEmpty() )
		{
			commandString += Application::windowsPath(file).replace("\\\\", "\\");
		}

		arguments << "-s" << commandString;

		if ( ! keyboardLayout.isEmpty() )
		{
			arguments << "-k" << keyboardLayout;
		}

		arguments += rdesktopOptions;
		arguments << server;
	}

	if ( QString( RDP_CLIENT ) == "mstsc.exe" )
	{
	}

	return arguments;
}

QString Application::cryptPassword(QString password)
{
	QString ret;
#ifdef Q_OS_LINUX
	ret = QCryptographicHash::hash( password.toUtf8(), QCryptographicHash::Md5);
#endif
#ifdef Q_OS_WIN
	DATA_BLOB DataIn;
	DATA_BLOB DataOut;
	BYTE * pbDataInput = (BYTE*)malloc(password.length() * sizeof(WCHAR));
	password.toWCharArray((wchar_t*)pbDataInput);
	DWORD cbDataInput = password.length() * sizeof(WCHAR);

	DataIn.pbData = pbDataInput;
	DataIn.cbData = cbDataInput;

	if(CryptProtectData( &DataIn, L"psw", nullptr, nullptr, nullptr, 0, &DataOut))
	{
		char f[2];
		for(uint i = 0 ; i < DataOut.cbData ; ++i)
		{
			sprintf(f, "%x%x", DataOut.pbData[i] >> 4, DataOut.pbData[i] & 0xf);
			ret += f;
		}
	} else {
		ret = "Error: The encryption phase didn't work.";
	}

	LocalFree( DataOut.pbData );
	free(pbDataInput);
#endif
	return ret.toUpper();
}

void Application::clearMstsc(QString tempFile)
{
	if ( QFile::exists(tempFile) )
	{
		QFile file(tempFile);
		file.remove();
	}
	/*
	QString command = "cmdkey.exe";
	QStringList arguments;
	arguments += "/delete:TERMSRV/" + hostName; // + ":" + QString::number( connPort );
	qDebug("Removing credentials: %s", QString( command + " " + arguments.join(" ")).toUtf8().data());
	int res = QProcess::execute(command, arguments);
	if ( res < 0 )
		QMessageBox::critical( 0, tr("Credentials"), tr("Unable to delete credentials") );
	*/
}

QStringList Application::mstscArguments(QString command, QStringList args, QString file, quint16 port, QString serverAddress)
{
	QStringList arguments;

	QTemporaryFile rdpTempFile;

	if ( rdpTempFile.open() )
	{
		QTextStream out( &rdpTempFile );

		// Non-configurable settings
		out << "compression:i:1\r\n";
		out << "keyboardhook:i:2\r\n";
		out << "redirectclipboard:i:1\r\n";
		out << "allow font smoothing:i:1\r\n";
		out << "authentication level:i:0\r\n";
		out << "prompt for credentials:i:0\r\n";
		out << "prompt for credentials on client:i:0\r\n";
		out << "promptcredentialonce:i:0\r\n";
		out << "use redirection server name:i:0\r\n";
		out << "displayconnectionbar:i:1\r\n";
		out << "pinconnectionbar:i:0\r\n";
		out << "autoreconnection enabled:i:1\r\n";
		out << "bitmapcachepersistenable:i:1\r\n";
		out << "bitmapcachesize:i:32000\r\n";
		out << "videoplaybackmode:i:1\r\n";
		out << "gatewayhostname:s:\r\n";
		out << "gatewayusagemethod:i:4\r\n";
		out << "gatewaycredentialssource:i:4\r\n";
		out << "gatewayprofileusagemethod:i:0\r\n";
		out << "gatewaybrokeringtype:i:0\r\n";
		out << "rdgiskdcproxy:i:0\r\n";
		out << "kdcproxyname:s:\r\n";

		// Configurable settings
		out << "full address:s:" + serverAddress + ":" + QString::number( port ) + "\r\n";
		out << "session bpp:i:32\r\n";
		/*QString connType = settings->getType( connName );
		if ( ! connType.isEmpty() )
		{
			if (connType == "RDP")
				out << "negotiate security layer:i:0\r\n";
			if (connType == "RDP TLS")
				out << "negotiate security layer:i:0\r\n";
			if (connType == "RDP NLA")
				out << "negotiate security layer:i:1\r\n";
		}*/

		out << "username:s:" + Settings::getUserName() + "\r\n";
		//out << "domain:s:" + Settings::getDomain( connName ) + "\r\n";
		out << "password 51:b:" + cryptPassword( Settings::getPassword() ) + "\r\n";
		/*
		// Redirect drives and dynamic devices
		out << "drivestoredirect:s:";
		if ( settings->getRemovableRedirect(connName))
		{
			out << "DynamicDevices;";
		}
		QStringList driveList = settings->getDriveRedirectFolders( connName );
		if (driveList.count() > 0)
		{

			for (int i = 0; i < driveList.count(); i++)
			{
				QString drive = driveList.at(i);
				out << drive.remove("\\") + ";";
			}
		}
		out << "\r\n";*/
		// End of redirect drives and dynamic devices
		out << "audiomode:i:0\r\n";
		out << "audiocapturemode:i:1\r\n";
		out << "redirectsmartcards:i:1\r\n";
		out << "redirectprinters:i:1\r\n";
		out << "redirectcomports:i:0\r\n";
		out << "redirectposdevices:i:0\r\n";
		out << "disable wallpaper:i:0\r\n";
		out << "allow desktop composition:i:0\r\n";
		out << "disable menu anims:i:1\r\n";
		out << "disable themes:i:1\r\n";
		out << "disable full window drag:i:1\r\n";
		out << "disable cursor setting:i:0\r\n";
		out << "remoteapplicationmode:i:1\r\n";
		out << "remoteapplicationprogram:s:" + command + " " + args.join(" ") + " " + file + "\r\n";
		out << "alternate shell:s:\r\n";
		out << "shell working directory:s:\r\n";
		rdpTempFile.close();
	} else {
		return QStringList();
	}

	rdpTempFile.copy( rdpTempFile.fileName() + ".rdp" );

	arguments += rdpTempFile.fileName() + ".rdp";

	return arguments;
}

QStringList Application::parseArguments( QString args )
{
	QStringList ret;
	bool insideQuote = false;
	bool insideSingleQuote = false;

	QString word;
	int i = 0;
	while ( i < args.size() )
	{
		if ( args.at(i) != '"' && args.at(i) != '\'' && args.at(i) != ' ') // Normal character
		{
			word += args.at(i);
			if (i == args.size() - 1) // last character
			{
				ret += word;
			}
		}
		if ( args.at(i) == '"' )
		{
			if (insideSingleQuote) // Cannot understand nested quotes, malformed quotation
			{
				i = args.size();
			}
			if (insideQuote) // End of quoted string
			{
				ret += word;
				word = QString();
				insideQuote = false;
			} else { // Begin of quoted string
				insideQuote = true;
			}
		}
		if ( args.at(i) == '\'' )
		{
			if (insideQuote) // Cannot understand nested quotes, malformed quotation
			{
				i = args.size();
			}
			if (insideSingleQuote) // End of single quoted string
			{
				ret += word;
				word = QString();
				insideSingleQuote = false;
			} else { // Begin of single quoted string
				insideSingleQuote = true;
			}
		}
		if ( args.at(i) == ' ' )
		{
			if (insideQuote || insideSingleQuote) // Space inside quotes
			{
				word += args.at(i);
			} else { // End of word
				ret += word;
				word = QString();
			}
		}
		i++;
	}
	if ( insideQuote || insideSingleQuote ) // malformed quotation
	{
		return args.split(' ');
	} else {
		return ret;
	}
}

QString Application::checkCommand(QString command, QString server, int os, QString protocol)
{
	bool valid = false;

	if ( ! command.contains(QDir::separator()) ) // command without path
	{
#ifdef Q_OS_WIN
		QStringList path = QString(qgetenv("PATH")).split(';');
#else
		QStringList path = QString(qgetenv("PATH")).split(':');
#endif
		for (int p = 0; p < path.count(); p++)
		{

			//qDebug("Checking: %s", QString( path[p] + QDir::separator() + command).toUtf8().data());
			if ( isValid( path[p] + QDir::separator() + command, server, os, protocol))
			{
				valid = true;
				break;
			}
		}
	}
#ifdef Q_OS_WIN
	command = resolveWindowsEnvironment( command );
#endif
	if ( ! valid )
	{
		return QString();
	} else {
		return command;
	}
}

void Application::setServerValues(QString id, Application * app)
{
	if ( LWClient::instance()->isConnected() )
	{
		app->setId( id );
		app->setServerId( LWClient::instance()->sendMessage( "get_preferred_server?" + id + "&0.5" ) );
		QByteArray ic = LWClient::instance()->sendMessageByteArray( "get_application_icon?" + id );
		QString server = LWClient::instance()->sendMessage( "get_server_address?" + app->getServerId() );
		app->setServer( server );
		int os = LINUX;
		QString type = LWClient::instance()->sendMessage( "get_server_type?" + app->getServerId() );
		if ( type == "Windows" )
		{
			os = WINDOWS;
		}
		if ( type == "Linux" )
		{
			os = LINUX;
		}
		app->setOS( os );
		QString protocol = "rdp";
		app->setProtocol( protocol );
		QString command = LWClient::instance()->sendMessage( "get_application_exec_path?" + id );
		app->setCommand( command );
		app->setArguments( parseArguments( LWClient::instance()->sendMessage( "get_application_arguments?" + id ) ) );
		app->setWorkingPath( LWClient::instance()->sendMessage( "get_application_working_path?" + id ) );
		app->setName( LWClient::instance()->sendMessage( "get_application_name?" + id ) );
		app->setDescription( LWClient::instance()->sendMessage( "get_application_display_name?" + id ) );
		QPixmap pixmap;
		pixmap.loadFromData( ic );
		app->setIcon( QIcon( pixmap ) );
	}
}

void Application::setLocalValues(QString id, Application * app)
{
	app->setId( id );
	QString command = checkCommand(  Settings::getApplicationValue( id, "command" ), "localhost", RUNNING_OS, QString() );
	app->setCommand( command );
	app->setArguments( Settings::getApplicationValue( id, "arguments" ).split(" ") );
	app->setWorkingPath( Settings::homePath() );
	app->setOS( Settings::getApplicationValue( id, "os" ).toInt() );
	app->setName( Settings::getApplicationValue( id, "name" ) );
	app->setDescription( Settings::getApplicationValue( id, "name" ) );
	QPixmap pixmap( ICON_DIR_APPS + QString(QDir::separator()) + Settings::getApplicationValue( id, "icon" ) );
	app->setIcon( QIcon( pixmap ) );
	app->setProtocol( QString() );
	// Get icon from .exe
	// this->icon = WinInfo::getAppIcon( command );
}

QString Application::firstString( QStringList stringList )
{
	if ( stringList.count() >= 1 )
		return stringList.at(0);
	else
		return QString();
}

void Application::launchFileManager(QString path)
{
	Application * getFileManager = Application::getFileManager();
	if ( getFileManager != nullptr )
	{
		Application * fileManager = new Application( getFileManager );
		fileManager->setFile( path );

		QStringList _arguments = fileManager->getArguments();

		QStringList fileLaunchParameters;
		fileLaunchParameters += Settings::getUserName();
		QString passwd = Settings::getPassword();
		if ( passwd.isEmpty() )
		{
			if (QFile::exists( TOKEN_FILE ))
			{
				passwd = Settings::getValue( TOKEN_FILE, "UserToken" );
			}
		}
		fileLaunchParameters += passwd;
		fileLaunchParameters += Settings::getServerAddress();
		fileLaunchParameters += QString::number( Settings::getServerPort() );
		if ( fileManager->getCommand() == "liberus-desktop" )
		{
			QTemporaryFile authFile;
			authFile.setAutoRemove( false );
			if ( authFile.open() )
			{
				for ( int i = 0; i < fileLaunchParameters.count(); i++ )
				{
					authFile.write(QString(fileLaunchParameters[i] + "\n").toUtf8());
				}
				authFile.close();
				_arguments += "-a";
				_arguments += authFile.fileName();
			}
		}

		fileManager->setArguments( _arguments );

		fileManager->launch();
		// fileManager->launch( fileLaunchParameters );
	}
}

void Application::launch( QStringList stdInput )
{
	QString exec;
	QStringList args;

	QProcess * process = new QProcess();

	QProcessEnvironment env = QProcessEnvironment::systemEnvironment();

	Settings::logDebug( QString("Server: %1, os: %2, isRemote: %3").arg(server).arg(Settings::getOsName(os)).arg( Application::isRemote(server, os)));

	if ( ( ! server.isEmpty() ) && Application::isRemote(server, os) )
	{
		exec = RDP_CLIENT;
#ifdef Q_OS_LINUX
		QProcess killXFreeRDP;
		killXFreeRDP.start("killall", QStringList("xfreerdp"));
		killXFreeRDP.waitForStarted();
		killXFreeRDP.waitForFinished();
#endif
	} else {
		if ( Settings::getBooleanValue( "ChrootEnvironment" ) )
		{
#ifdef Q_OS_LINUX
			exec = CHROOT;
#endif
		} else {
#ifdef Q_OS_WIN
			exec = SHAREDIR "liberuslauncher.bat";
#else
			exec = "liberuslauncher.sh";
#endif
		}
		env.insert("LANG", "es_ES.UTF-8");
		//env.insert("LANG", "de_DE.UTF-8");
		//env.insert("LANG", "fr_FR.UTF-8");
	}
	process->setProcessEnvironment(env);

	qDebug("Command: '%s'", command.toUtf8().data());
	args += Application::getArguments( command, arguments, file, os, server );

	Settings::logInfo( QString ("Launching: '%1 %2'").arg(exec).arg( args.join(" ")) );

	if ( workingPath.simplified().isEmpty() )
	{
		workingPath = Settings::homePath();
	}
	Settings::logInfo( QString ("Working path: %1").arg( workingPath.toUtf8().data() ) );
		process->setWorkingDirectory( workingPath );

	if ( stdInput.count() > 0 )
	{
		process->start( exec, args );
		if ( process->waitForStarted() )
		{
			pid = process->processId();
			for ( int i = 0; i < stdInput.count(); i++ )
			{
				QString output = stdInput[i] + "\n";
				process->write( output.toUtf8() );
			}
			connect(process, SIGNAL(finished(int, QProcess::ExitStatus)), this, SLOT(finish(int, QProcess::ExitStatus)));
		}
	} else {
		if ( process->startDetached( exec, args, QString(),(qint64*) &pid ) )
		{
			connect(&fileSystemWatcher, SIGNAL(fileChanged(QString)), this, SLOT(watcherFileChanged(QString)));
		}
	}

	if (LWClient::instance()->isConnected())
	{
		QString res = LWClient::instance()->sendMessage( "start_application?" + id.trimmed() + "&" + QString::number((unsigned long long)pid).trimmed() );
		Settings::logInfo( "start_application?" + id.trimmed() + "&" + QString::number((unsigned long long)pid).trimmed() + " with result: " + res );
	}
}

void Application::watcherFileChanged( QString fileName )
{
	if ( pid != 0 )
	{
		if ( QFileInfo( fileName ).baseName() == QString::number( reinterpret_cast<long long>(pid) ) )
		{
			finish( 0, QProcess::NormalExit );
		}
	}
}

void Application::finish( int exitCode, QProcess::ExitStatus exitStatus )
{
	if ( LWClient::instance()->isConnected() )
	{
		QString res = LWClient::instance()->sendMessage( "end_application?" + id.trimmed() + "&" + QString::number((unsigned long long)pid).trimmed() );
		Settings::logInfo( "Sending end_application?" + id.trimmed() + "&" + QString::number((unsigned long long)pid).trimmed() + " with result: " + res );
	}

	emit finished( exitCode, exitStatus );
}

void Application::deleteAppInstance()
{
	delete this;
}
