#include "filethread.h"

FileThread::FileThread(FileOperation operation, QObject *parent, QFileInfo sourceFile, QDir targetDir)
{
	_operation = operation;
	_sourceFile = sourceFile;
	_sourceFiles.append(_sourceFile);
	FileThread( operation, parent, _sourceFiles, targetDir );
}

FileThread::FileThread(FileOperation operation, QObject *parent, QFileInfoList sourceFiles, QDir targetDir) : QThread( parent )
{
	_operation = operation;
	_sourceFiles = sourceFiles;
	_targetDir = targetDir;
	_terminate = false;
	_pause = false;

	askToReplace = true;
	replace = false;

	QString targetRootPath = QStorageInfo( targetDir ).rootPath();
	_isSameFileSystem = true;
	for ( int i = 0; i < sourceFiles.count(); i++ )
	{
		QString sourceRootPath = QStorageInfo( sourceFiles[i].absoluteFilePath() ).rootPath();
		if ( sourceRootPath != targetRootPath )
		{
			_isSameFileSystem = false;
		}
	}
}

FileThread::FileThread(FileOperation operation, QObject *parent, QStringList sourceFiles, QString targetDir) : QThread( parent )
{
	_operation = operation;
	_sourceFilesStringList = sourceFiles;
	_targetDirString = targetDir;
	_terminate = false;
	_pause = false;

	askToReplace = true;
	replace = false;

	QString targetRootPath = QStorageInfo( QDir(targetDir) ).rootPath();
	_isSameFileSystem = true;
	for ( int i = 0; i < sourceFiles.count(); i++ )
	{
		QString sourceRootPath = QStorageInfo( sourceFiles[i] ).rootPath();
		if ( sourceRootPath != targetRootPath )
		{
			_isSameFileSystem = false;
		}
	}
}

void FileThread::run()
{
	_sizes = 0;
	_copied = 0;
	_files = 0;
	_dirs = 0;

	finishMutex.lock();
	if ( _operation == REMOVE_FILES )
	{
		for ( int i = 0; i < _sourceFiles.count(); i++ )
		{
			emit currentFile( _sourceFiles[i].fileName() );
			removeFile( _sourceFiles[i] );
		}
		emit( finished() );
		finishMutex.unlock();
		return;
	}
	if ( _operation == REMOTE_COPY )
	{
		for ( int i = 0; i < _sourceFilesStringList.count(); i++ )
		{
			getDirSize( QFileInfo( _sourceFilesStringList[i] ) );
		}
	}
	if ( _operation == CALCULATE_SIZES || _operation == COPY_FILES || ( _operation == MOVE_FILES && ! _isSameFileSystem ) )
	{
		emit changedCurrentOperation( CALCULATE_SIZES );
		for ( int i = 0; i < _sourceFiles.count(); i++ )
		{
			getDirSize( _sourceFiles[i] );
		}

		if ( _operation == COPY_FILES || _operation == MOVE_FILES )
		{
			QStorageInfo targetDirInfo( _targetDir );
			long long bytesAvailable = 0;
#ifdef Q_OS_WIN
			ULARGE_INTEGER lpFreeBytesAvailableToCaller, lpTotalNumberOfBytes, lpTotalNumberOfFreeBytes;
			BOOL getSpaceOk = GetDiskFreeSpaceEx((LPCWSTR) _targetDir.dirName().utf16(), &lpFreeBytesAvailableToCaller, &lpTotalNumberOfBytes, &lpTotalNumberOfFreeBytes);
			if ( getSpaceOk )
			{
				bytesAvailable = (long long)lpTotalNumberOfFreeBytes.QuadPart;
			}
#else
			bytesAvailable = targetDirInfo.bytesAvailable();
#endif
			if ( _sizes > bytesAvailable && false)
			{
				emit error(FileThread::NOT_ENOUGH_SPACE, QStringList(_targetDir.absolutePath()));
				finishMutex.unlock();
				mutex.lock();
				_terminate = true;
				mutex.unlock();
				return;
			}
		}
	}
	if ( _operation == COPY_FILES )
	{
		emit changedCurrentOperation( COPY_FILES );
		for ( int i = 0; i < _sourceFiles.count(); i++ )
		{
			copyFile( _sourceFiles[i], _targetDir );
		}
	}
	if ( _operation == REMOTE_COPY )
	{
		emit changedCurrentOperation( REMOTE_COPY );

		copyFilesToProfilesServer( _sourceFilesStringList, _targetDirString );
	}
	if ( _operation == MOVE_FILES )
	{
		emit changedCurrentOperation( MOVE_FILES );
		for ( int i = 0; i < _sourceFiles.count(); i++ )
		{
			moveFile( _sourceFiles[i], _targetDir );
		}
	}
	emit( finished() );
	finishMutex.unlock();
}

long long FileThread::sizes()
{
	return _sizes;
}

int FileThread::files()
{
	return _files;
}

int FileThread::dirs()
{
	return _dirs;
}

void FileThread::setOperation(FileOperation operation)
{
	this->_operation = operation;
}

FileThread::FileOperation FileThread::operation()
{
	return _operation;
}

void FileThread::finishThread()
{
	mutex.lock();
	_terminate = true;
	mutex.unlock();

	finishMutex.lock();
	quit();
	finishMutex.unlock();
}

void FileThread::setUserResponse(QMessageBox::StandardButton response)
{
	mutex.lock();
	if ( response == QMessageBox::YesToAll || response == QMessageBox::NoToAll )
	{
		askToReplace = false;
	}
	if ( response == QMessageBox::No || response == QMessageBox::NoToAll )
	{
		replace = false;
	}
	if ( response == QMessageBox::Yes || response == QMessageBox::YesToAll )
	{
		replace = true;
	}
	if ( response == QMessageBox::Cancel )
	{
		_terminate = true;
	}
	mutex.unlock();

	waitMutex.lock();
	waitUser.wakeAll();
	waitMutex.unlock();
}

void FileThread::waitUserResponse()
{
	waitMutex.lock();
	waitUser.wait( &waitMutex );
	waitMutex.unlock();
}

void FileThread::removeFile(QFileInfo file)
{
	QDir dir = file.absoluteDir();
	if ( file.isSymLink() )
	{
		dir.remove( file.fileName() );
	} else {
		if ( file.isDir() )
		{
			QDir fullDir = QDir( file.absoluteFilePath() );
			fullDir.removeRecursively();
		} else {
			dir.remove( file.fileName() );
		}
	}
}

void FileThread::moveFile(QFileInfo sourceFile, QDir targetDir)
{
	if ( _isSameFileSystem )
	{
		QDir dir( sourceFile.absoluteFilePath() );
		QString targetFileName = targetDir.absoluteFilePath( sourceFile.fileName() );
		if ( dir.rename( sourceFile.absoluteFilePath(), targetFileName ) )
		{
			emit copyProgress( sourceFile, 1, 1, 1, 1 );
		} else {
			mutex.lock();
			_terminate = true;
			mutex.unlock();
			QStringList entryList;
			entryList += dir.absolutePath();
			emit error( CANNOT_MOVE_FILES, entryList );
		}
	} else {
		copyFile( sourceFile, targetDir, true );
	}
}

void FileThread::copyFile(QFileInfo sourceFile, QDir targetDir, bool move)
{
	qDebug("sourceFile: '%s' targetDir: '%s'", sourceFile.absoluteFilePath().toUtf8().data(), targetDir.path().toUtf8().data());
	// Check for termination command
	mutex.lock();
	if ( _terminate )
	{
		mutex.unlock();
		return;
	} else {
		mutex.unlock();
	}

	QString targetFileName = targetDir.absoluteFilePath( sourceFile.fileName() );

	if ( askToReplace )
	{
		if ( QFile::exists( targetFileName ) )
		{
			emit error( FILE_EXISTS, QStringList(targetFileName) );
			waitUserResponse();
		}
	}

	// First, copy links
	if ( sourceFile.isSymLink() )
	{
		long long fileSize = sourceFile.size();
		emit copyProgress( sourceFile, _sizes, _copied, fileSize, 0 );
		//qDebug("_sizes: %d, _copied: %d, fileSize: %d, fileCopied: %d", _sizes, _copied, fileSize, 0);
		// TODO: create link on real file server
		/*if ( LWClient::isConnected() )
		{
			QString server = LWClient::sendMessage("get_storage_server_for_file", QStringList(targetFileName) );
			QStringList parameters;
			parameters += server;
			parameters += sourceFile.symLinkTarget();
			parameters += targetFileName;
			QString ret = LWClient::sendMessage("create_file_link", parameters );
		} else {
		*/
		QString linkCommand;
		QStringList linkArguments;
#ifdef Q_OS_LINUX
		linkCommand = "ln";
		linkArguments += "-s";
		linkArguments += sourceFile.symLinkTarget();
		linkArguments += targetFileName;
#endif
#ifdef Q_OS_WIN
		linkCommand = "MKLINK";
		if ( sourceFile.isDir() )
		{
			linkArguments += "/d";
		}
		linkArguments += targetFileName;
		linkArguments += sourceFile.symLinkTarget();
#endif
		QProcess::execute( linkCommand, linkArguments );
		_copied = _copied + fileSize;
		emit copyProgress( sourceFile, _sizes, _copied, fileSize, fileSize );
		return;
	}

	if ( sourceFile.isDir() )
	{
		long long fileSize = sourceFile.size();
		emit copyProgress( sourceFile, _sizes, _copied, fileSize, 0 );
		if ( ! targetDir.mkdir( sourceFile.fileName() ) )
		{
			emit error( CANNOT_CREATE_DIR, QStringList(sourceFile.fileName()) );
			mutex.lock();
			_terminate = true;
			mutex.unlock();
			return;
		} else {
			_copied = _copied + fileSize;
			emit copyProgress( sourceFile, _sizes, _copied, fileSize, fileSize );
		}

		QDir newTargetDir = QDir( targetFileName );
		QDir dir( sourceFile.absoluteFilePath() );
		QList<QFileInfo> children = dir.entryInfoList(QDir::Files | QDir::Dirs | QDir::NoDotAndDotDot | QDir::Hidden | QDir::System);
		for ( int i = 0; i < children.count(); i++ )
		{
			copyFile( children[i], newTargetDir, move);
		}
		if ( move )
		{
			QStringList entryList = dir.entryList(QDir::NoDotAndDotDot);
			if ( ! entryList.count() == 0 ) // dir not empty after moving contents
			{
				mutex.lock();
				_terminate = true;
				mutex.unlock();
				emit error( CANNOT_MOVE_FILES, entryList );
				return;
			} else {
				dir.removeRecursively();
			}
		}
	} else {
		if ( move )
		{
			emit copyProgress( sourceFile, _sizes, _copied, 0, 0);
			QDir dir( sourceFile.absoluteFilePath() );
			long long fileSize = sourceFile.size();
			if ( dir.rename( sourceFile.absoluteFilePath(), targetFileName ) )
			{
				_copied = _copied + fileSize;
				emit copyProgress( sourceFile, _sizes, _copied, fileSize, fileSize);
				return;
			} // else continue with normal copy, probably different filesystems so cannot rename
		}

		QFile file( sourceFile.absoluteFilePath() );
		QFileDevice::Permissions perm = file.permissions();
		file.open(QIODevice::ReadOnly);

		QFile targetFile( targetFileName );
		targetFile.open(QIODevice::WriteOnly);
		long long position = 0;
		long long bufSize = 1024 * 1024; // 1 MiB buffer
		long long fileSize = file.size();
		emit copyProgress( sourceFile, _sizes, _copied, fileSize, 0);
		for ( position = 0; position < fileSize; position = position + bufSize )
		{
			mutex.lock();
			if ( _terminate )
			{
				mutex.unlock();
				break;
			} else {
				mutex.unlock();
			}
			char * data = (char*)malloc( bufSize * sizeof(char) );

			long long bytesRead = file.read( data, bufSize );
			if ( bytesRead == -1 )
			{
				emit error( CANNOT_READ_FILE, QStringList(sourceFile.fileName()) );
				mutex.lock();
				_terminate = true;
				mutex.unlock();
			}
			long long bytesWritten = targetFile.write(data, bytesRead);
			if ( bytesWritten != bytesRead )
			{
				emit error( CANNOT_WRITE_FILE, QStringList(targetFileName) );
				mutex.lock();
				_terminate = true;
				mutex.unlock();
			}
			_copied = _copied + bytesWritten;
			free(data);

			emit copyProgress( sourceFile, _sizes, _copied, fileSize, position + bytesWritten);
		}
		file.close();
		targetFile.setPermissions( perm );
		targetFile.close();
		if ( move )
		{
			file.remove();
		}
	}
}

void FileThread::copyFilesToProfilesServer(QStringList fileNames, QString targetDir)
{
	QString token = Settings::getValue( TOKEN_FILE, "UserToken" );
	QString profilesServerAddress = Settings::getProfilesServerAddress();
	int profilesServerPort = Settings::getProfilesServerPort();
	qDebug("Profiles server: %s:%d", profilesServerAddress.toUtf8().data(), profilesServerPort);

	long long totalCopied = 0;
	for ( int i = 0; i < fileNames.count(); i++ )
	{
		LWClient * lwClient = new LWClient();
		if ( lwClient->connect( profilesServerAddress, profilesServerPort ) )
		{
			bool auth = false;
			if ( ! token.isEmpty() )
			{
				auth = lwClient->authenticate( Settings::getFullUserName(), token );
			} else {
				qDebug("User token to connect to server not found");
			}
			if ( auth )
			{
				QString fileName = fileNames[i];
				QFile file( fileName );
				QFileInfo fileInfo( fileName );
				long long fileSize = file.size();

				emit copyProgress( fileInfo, _sizes, totalCopied, fileSize, 0);
				QString command = "_sendBinaryFile_";

				QString message = command + "?"
							+ fileInfo.fileName() + "&"
							+ targetDir + "&"
							+ QString::number( fileSize );
				qDebug("Sending message: '%s'", message.toUtf8().data());
				QString response = lwClient->sendMessage( message );
				qDebug("Response: %s", response.toUtf8().data());
				qDebug("Copying file: '%s'", fileName.toUtf8().data());
				//LWSocket::setSocketBuffer( lwClient->socket(), 65536 );
				//int flags = 1;
				//setsockopt(lwClient->socket(), SOL_SOCKET, TCP_NODELAY, (void *)&flags, sizeof(flags));
				if ( file.open(QIODevice::ReadOnly) )
				{
					QByteArray buffer;
					long long numBytes = 1;
					long long totalBytes = 0;
					while ( ! file.atEnd() && numBytes > 0 )
					{
						buffer = file.read( FILE_SEND_BUFFER_SIZE );
						numBytes = lwClient->sendData( buffer );
						totalBytes += numBytes;
						totalCopied += numBytes;
						emit copyProgress( fileInfo, _sizes, totalCopied, fileSize, totalBytes);
						mutex.lock();
						if ( _terminate )
						{
							mutex.unlock();
							break;
						} else {
							mutex.unlock();
						}
					}
					qDebug("Copied %lld bytes", totalBytes);
					file.close();
				} else {
					qDebug("Cannot open file '%s' for reading", fileName.toUtf8().data() );
				}
			} else {
				qDebug("Cannot connect with current user token");
			}
			lwClient->close(false);
		} else {
			qDebug("Cannot connect to Liberus Server");
		}
		delete lwClient;
	}
	emit finished( fileNames, targetDir );
}

void FileThread::getDirSize(QFileInfo file)
{
	bool term;
	mutex.lock();
	term = _terminate;
	mutex.unlock();
	if ( ! term )
	{
		_sizes = _sizes + file.size();
		if ( file.isDir() && ! file.isSymLink() )
		{
			_dirs++;
			QDir dir(file.absoluteFilePath());
			QList<QFileInfo> children = dir.entryInfoList(QDir::Files | QDir::Dirs | QDir::NoDotAndDotDot | QDir::Hidden | QDir::System);
			for ( int i = 0; i < children.count(); i++ )
			{
				if ( ! children[i].isSymLink() )
				{
					mutex.lock();
					term = _terminate;
					mutex.unlock();
					if ( ! term )
					{
						getDirSize( children[i] );
					}
				}
			}
		} else {
			_files++;
		}
		emit calculateProgress( _sizes, _files, _dirs );
	}
}
