#include "filethreadwidget.h"

FileThreadWidget::FileThreadWidget(FileThread * fileThread, QWidget *parent) : QDialog( parent )
{
	_close = false;
	setWindowModality( Qt::WindowModal );

	if ( fileThread->operation() == FileThread::COPY_FILES )
	{
		setWindowTitle( tr("Copy") );
		setWindowIcon(QIcon(":/icons/64x64/copy.png"));
	}
	if ( fileThread->operation() == FileThread::REMOTE_COPY )
	{
		setWindowTitle( tr("Remote copy") );
		setWindowIcon(QIcon(":/icons/64x64/copy.png"));
	}
	if ( fileThread->operation() == FileThread::MOVE_FILES )
	{
		setWindowTitle( tr("Copy") );
		setWindowIcon(QIcon(":/icons/64x64/rename.png"));
	}
	if ( fileThread->operation() == FileThread::REMOVE_FILES )
	{
		setWindowTitle( tr("Remove") );
		setWindowIcon(QIcon(":/icons/64x64/remove.png"));
		setMinimumWidth( 400 );
	}
	timer = new QTimer();
	timer->setTimerType(Qt::VeryCoarseTimer);
	this->fileThread = fileThread;

	maxSpeed = 0;

	initUi();
	qRegisterMetaType<FileThread::FileError>();
	qRegisterMetaType<FileThread::FileOperation>();
	connect( fileThread, SIGNAL(changedCurrentOperation(FileThread::FileOperation)), this, SLOT(changedCurrentOperation(FileThread::FileOperation)));
}

FileThreadWidget::~FileThreadWidget()
{
	timer->stop();
	delete timer;
	delete fileThread;
}

void FileThreadWidget::initUi()
{
	QGridLayout * layout = new QGridLayout;
	layout->setContentsMargins( DEFAULT_MARGIN, DEFAULT_MARGIN, DEFAULT_MARGIN, DEFAULT_MARGIN );
	layout->setSpacing( DEFAULT_SPACING );

	fileName = new QLabel;
	partialProgressBar = new QProgressBar;
	partialProgressBar->setValue( 0 );
	partialProgressBar->setFixedHeight( 16 );
	partialProgressBar->setFixedWidth( 150 );
	partialProgressBar->setFormat( " %p%" );

	progressBar = new QProgressBar;
	progressBar->setValue( 0 );
	progressBar->setFixedHeight( 16 );
	progressBar->setFormat( " %p%" );

	status = new QLabel;

	speed = new QLabel;

	acceptButton = new QPushButton( tr("Accept"), this);
	acceptButton->setFlat( true );
	acceptButton->setStyleSheet(STANDARDBUTTONSTYLE);
	acceptButton->setMinimumSize( 100, 24 );
	acceptButton->setVisible( false );
	acceptButton->setFixedWidth( 100 );
	connect( acceptButton, SIGNAL(clicked(bool)), this, SLOT(close()));

	cancelButton = new QPushButton( tr("Cancel"), this);
	cancelButton->setFlat( true );
	cancelButton->setStyleSheet(STANDARDBUTTONSTYLE);
	cancelButton->setMinimumSize( 100, 24 );
	cancelButton->setFixedWidth( 100 );
	connect( cancelButton, SIGNAL(clicked(bool)), this, SLOT(cancel()));

	layout->addWidget(fileName, 0, 0, 1, 4);
	layout->addWidget(partialProgressBar, 0, 5, 1, 1);
	layout->addWidget(progressBar, 1, 0, 1, 6);
	layout->addWidget(status, 2, 0, 1, 6);
	layout->addWidget(speed, 3, 0, 1, 6);
	layout->addWidget(acceptButton, 4, 4);
	layout->addWidget(cancelButton, 4, 5);

	setLayout( layout );
	setStyleSheet(QString( STANDARDDIALOGSTYLE ));
}

void FileThreadWidget::cancel()
{
	fileThread->finishThread();
	_close = true;
	close();
}

void FileThreadWidget::changedCurrentOperation(FileThread::FileOperation operation)
{
	if ( operation == FileThread::CALCULATE_SIZES )
	{
		fileName->setText( tr("Calculating size...") );
		partialProgressBar->setVisible( false );
		progressBar->setRange(0, 0);
	}
	if ( operation == FileThread::COPY_FILES || operation == FileThread::MOVE_FILES || operation == FileThread::REMOTE_COPY )
	{
		partialProgressBar->setVisible( true );
	}
}

void FileThreadWidget::closeEvent(QCloseEvent * event)
{
	if ( _close )
	{
		QDialog::closeEvent( event );
	} else {
		event->ignore();
	}
}

void FileThreadWidget::start()
{
	if ( fileThread->operation() == FileThread::REMOVE_FILES )
	{
		partialProgressBar->hide();
		speed->hide();
		progressBar->setRange(0, 0);
		cancelButton->hide();
		connect( fileThread, SIGNAL(currentFile(QString)), this, SLOT(setCurrentFile(QString)));
	}
	if ( fileThread->operation() == FileThread::COPY_FILES || fileThread->operation() == FileThread::MOVE_FILES || fileThread->operation() == FileThread::REMOTE_COPY )
	{
		connect( fileThread, SIGNAL(copyProgress(QFileInfo,long long,long long,long long,long long)), this, SLOT(updateValues(QFileInfo,long long,long long,long long,long long)));
		connect( fileThread, SIGNAL(error(FileThread::FileError,QStringList)), this, SLOT(error(FileThread::FileError,QStringList)));
	}
	if ( fileThread->operation() == FileThread::CALCULATE_SIZES )
	{
		connect( timer, SIGNAL(timeout()), this, SLOT(updateValues()));
		timer->start( 1000 );
	}
	connect( fileThread, SIGNAL(finished()), this, SLOT(finished()));
	startTime = QDateTime::currentDateTime();
	fileThread->start();
}

void FileThreadWidget::error(FileThread::FileError error, QStringList files)
{
	if ( files.count() == 0 )
	{
		QMessageBox::critical( this, tr("Unknown error"), tr("An unknown error happened"), QMessageBox::Ok);
		close();
		deleteLater();
		return;
	}
	if ( error == FileThread::CANNOT_CREATE_DIR )
	{
		QMessageBox::critical( this, tr("Cannot create folder"), tr("Error: Cannot create '%1' folder.").arg(files[0]), QMessageBox::Ok);
	}
	if ( error == FileThread::CANNOT_READ_FILE )
	{
		QMessageBox::critical( this, tr("Cannot read file"), tr("Error: Cannot read '%1' file.").arg(files[0]), QMessageBox::Ok);
	}
	if ( error == FileThread::CANNOT_WRITE_FILE )
	{
		QMessageBox::critical( this, tr("Cannot write file"), tr("Error: Cannot write '%1' file.").arg(files[0]), QMessageBox::Ok);
	}
	if ( error == FileThread::CANNOT_MOVE_FILES )
	{
		QMessageBox::critical( this, tr("Cannot move files"), tr("Error: Cannot move files:\n'%1'").arg(files.join("\n")), QMessageBox::Ok);
	}
	if ( error == FileThread::NOT_ENOUGH_SPACE )
	{
		QMessageBox::critical( this, tr("Not enough space"), tr("Error: Target folder %1 doesn't have enough free space").arg(files[0]), QMessageBox::Ok);
	}
	if ( error == FileThread::INSUFFICIENT_PERMISSIONS )
	{
		QMessageBox::critical( this, tr("Insufficient permissions"), tr("Error: Not enough permissions on file '%1'").arg(files[0]), QMessageBox::Ok);
	}
	if ( error == FileThread::FILE_EXISTS )
	{
		QMessageBox::StandardButton ret = QMessageBox::warning( this, tr("File exists"), tr("File or folder '%1' currently exists. Do you want to overwirte it?").arg(files[0]), QMessageBox::Yes | QMessageBox::No | QMessageBox::YesToAll | QMessageBox::NoToAll | QMessageBox::Cancel, QMessageBox::No);
		fileThread->setUserResponse( ret );
		return; // Don't close or delete this widget
	}
	// Close and delete for all errors (except warnings which returned previously)
	close();
	deleteLater();
}

void FileThreadWidget::setCurrentFile(QString file)
{
	if ( fileThread->operation() == FileThread::REMOVE_FILES )
	{
		fileName->setText( tr("Removing %1...").arg(file) );
	} else {
		fileName->setText( file );
	}
}

void FileThreadWidget::updateValues()
{
	long long seconds = startTime.secsTo(QDateTime::currentDateTime());
	QString statusString = tr("Elapsed time:") + readableSeconds(seconds);
	status->setText( statusString );
}

void FileThreadWidget::updateValues(QFileInfo file, long long totalSize, long long totalCopied, long long currentFileSize, long long currentFileCopied)
{
	fileName->setText( file.fileName() );
	if ( currentFileSize > 0 && totalSize > 0 )
	{
		partialProgressBar->setRange( 0, 100 );
		partialProgressBar->setValue( ( 100 * currentFileCopied ) / currentFileSize );
		progressBar->setRange( 0, 100 );
		progressBar->setValue( ( 100 * totalCopied ) / totalSize );

		long long elapsedTime = startTime.secsTo(QDateTime::currentDateTime());
		long long percentageCopied = ( 100 * totalCopied ) / totalSize;

		QStringList statusStringList;
		statusStringList += FileProperties::toHuman( totalCopied ) + "/" + FileProperties::toHuman( totalSize );

		if ( elapsedTime > 0 )
		{
			long long speedNumber = totalCopied / elapsedTime;
			speed->setText( FileProperties::toHuman( speedNumber ) + "/s" );
		}

		long long remainingTime;
		if ( percentageCopied != 0 )
		{
			remainingTime = ( ( ( 100 - percentageCopied ) * elapsedTime ) / percentageCopied );
			statusStringList += "(" + readableSeconds(remainingTime) + " " + tr("remaining") +")";
		}
		status->setText( statusStringList.join(" "));
	}
}

void FileThreadWidget::finished()
{
	timer->stop();
	acceptButton->setVisible( true );
	cancelButton->setVisible( false );
	status->setText(tr("Finished."));
	_close = true;
	if (	fileThread->operation() == FileThread::REMOVE_FILES ||
			fileThread->operation() == FileThread::COPY_FILES ||
			fileThread->operation() == FileThread::MOVE_FILES ||
			fileThread->operation() == FileThread::REMOTE_COPY )
	{
		close();
		deleteLater();
	}
}

QString FileThreadWidget::readableSeconds( long long seconds )
{
	if ( seconds <= 120 ) // two minutes
	{
		if ( seconds == 1 )
		{
			return QString::number(seconds) + " " + tr("second");
		} else {
			return QString::number(seconds) + " " + tr("seconds");
		}
	} else if ( seconds <= 120 * 60 ) { // two hours
		QString number = QString::number( seconds / 60 );
		if ( number == "1" )
		{
			return number + " " + tr("minute");
		} else {
			return number + " " + tr("minutes");
		}
	} else {
		QString ret;
		long long hours = seconds / 3600;
		long long minutes = seconds - hours * 3600;
		QString numberHours = QString::number( hours );
		if ( numberHours == "1" )
		{
			ret = numberHours + " " + tr("hour");
		} else {
			ret = numberHours + " " + tr("hours");
		}
		if ( minutes > 0 )
		{
			ret += " " + tr("and") + " ";
			if ( minutes == 1 )
			{
				ret += "1 " + tr("minute");
			} else {
				ret += QString::number( minutes ) + " " + tr("minutes");
			}
		}
		return ret;
	}
}
