#include "manager.h"

QProcess * Manager::wmProcess = nullptr;

Manager::Manager(InitProgress *initProgress, QString desktopUsername, QString desktopPassword, QString serverAddress, int serverPort, QWidget * parent)
{
	/*Settings::logInfo( "Starting window manager..." );
	startWindowManager(); // start configured external window manager (openbox, sawfish, ...)
	Settings::logInfo( "Done." );*/

	QFont font = QFont( Settings::getUserValue("Font", LIBERUS_FONT).toString(), Settings::getUserValue("FontSize", QString::number(LIBERUS_FONT_BASE_SIZE)).toInt());
	qApp->setFont( font );

	this->parent = parent;
	if ( ! desktopUsername.isEmpty() )
	{
		Settings::setUserName( desktopUsername );
	}
	if ( ! desktopPassword.isEmpty() )
	{
		Settings::setPassword( desktopPassword );
	}
	Settings::setServerAddress( serverAddress );
	Settings::setServerPort( serverPort );
	this->initProgress = initProgress;
	this->serverAddress = serverAddress;
	this->serverPort = serverPort;
}

void Manager::init()
{
	createDesktopLinks( QDir::homePath() );

	launchingRemote = QString();
	appFileWatcher = nullptr;
	QString username = Settings::getUserName();
	Settings::logInfo( QString( "Starting session with user '%1'" ).arg(username) );

	qDebug("Settings file: %s", QString(CONFIG_FILE).toUtf8().data());

	if ( ! serverAddress.isEmpty() && serverPort > 0 )
	{
		QString token;
		qDebug("Looking for token on file: %s", QString(TOKEN_FILE).toUtf8().data());
		if (QFile::exists( TOKEN_FILE ))
		{
			token = Settings::getValue( TOKEN_FILE, "UserToken" );
		}

		if (connectToServer( serverAddress, serverPort ))
		{
			bool auth = false;
			QString password = Settings::getPassword();
			if ( ! password.isEmpty() )
			{
				auth = LWClient::instance()->authenticate( username, password );
			} else {
				if ( ! token.isEmpty() )
				{
					auth = LWClient::instance()->authenticate( username, token );
				}
			}
			if ( auth )
			{
				qDebug("Authentication ok");
			}
			if ( ! auth )
			{
				InitScreen * initScreen = new InitScreen();
				bool auth = false;

				while (! auth )
				{
					initScreen->setFixedSize(QApplication::primaryScreen()->size());
					initScreen->showFullScreen();
					initScreen->centerDialog();
					int res = initScreen->exec();
					if ( res == QDialog::Rejected )
					{
						qDebug("Rejecting...");
						delete initScreen;
						exitApp();
						return;
					} else {
						QString initPassword = initScreen->getPassword();
						Settings::setPassword( initPassword );
						auth = LWClient::instance()->authenticate( username, initPassword );
						if ( auth )
						{
							token = LWClient::instance()->sendMessage("activate_user?" + username);
							if (! token.isEmpty()  && ! token.startsWith("Error:") )
							{
								Settings::setValue(TOKEN_FILE, "UserToken", token);
							} else {
								QMessageBox::information(nullptr, tr("Server connection"), tr("Cannot enable user. Contact your system administrator."));
								exitApp();
								return;
							}
						} else {
							QMessageBox::information(nullptr, tr("Server connection"), tr("The password you supplied is incorrect. Please, try again."));
						}
					}
				}
			}
		} else {
			QMessageBox::information(nullptr, tr("Server connection"), tr("Cannot connect to Liberus Server. Contact your system administrator."));
			exitApp();
			return;
		}
		QString res = LWClient::instance()->sendMessage( "start_session" );
		Settings::logInfo( res );
	} else {
#ifdef Q_OS_LINUX
		QString apps = "/usr/share/applications";
		QDir appsDir( apps );
		QStringList appsDirEntries = appsDir.entryList(QStringList(), QDir::AllDirs);
		QStringList watchDir;
		for ( int i = 0; i < appsDirEntries.count(); i++ )
		{
			if ( appsDirEntries[i] != ".." )
			{
				watchDir += apps + "/" + appsDirEntries[i];
			}
		}
		appFileWatcher = new QFileSystemWatcher(watchDir);
#endif
	}

	Application::deleteApplicationCache();
	Application::updateApplicationCache( initProgress );

	connect( QApplication::primaryScreen(), SIGNAL( geometryChanged(const QRect &) ), this, SLOT( resize() ) );

#ifdef Q_OS_WIN
	sessionManager = new SessionManager();
	connect( sessionManager, SIGNAL(clientConnectEvent()), this, SLOT(clientConnected()));
	sessionManager->start();
#endif
	winManager = new WinManager( this );

	tabBar = new TabBar( initProgress, nullptr );

	winManager->appendOwnWindow( (HWND)tabBar->getDesktop()->winId() );
	winManager->appendOwnWindow( (HWND)tabBar->winId() );

	connect (tabBar, SIGNAL(launchApplication(Application*)), this, SLOT(launchApplication(Application*)) );
	connect( tabBar, SIGNAL(exitApp()), this, SLOT(exitApp()));
	connect( tabBar, SIGNAL(threadButtonClicked(HWND)), this, SLOT(mapWindow(HWND)) );
	connect( tabBar, SIGNAL(newThread()), this, SLOT(loadedFinished()) );
	connect( tabBar, SIGNAL(desktopButtonClicked()), winManager, SLOT(desktopRequested()) );

	if ( appFileWatcher != nullptr )
	{
		connect( appFileWatcher, SIGNAL(directoryChanged(QString)), tabBar, SLOT(updateApps()));
	}

	connect( tabBar->getDesktop(), SIGNAL(selectWidget(QWidget*)), this, SLOT(mapWidget(QWidget*)) );
	connect( tabBar->getDesktop(), SIGNAL(selectWidget(QWidget*)), tabBar->getCalendarWidget(), SLOT(hide()) );

	connect( winManager, SIGNAL(newApplication( HWND )), tabBar, SLOT(newWindow( HWND )));
	connect( winManager, SIGNAL(removeApplication( HWND )), tabBar, SLOT(removeWindow( HWND )));
	connect( winManager, SIGNAL(iconChanged( HWND, QPixmap )), tabBar, SLOT(changeIcon( HWND, QPixmap )));
	connect( winManager, SIGNAL(setSelected( HWND )), tabBar, SLOT(setSelected( HWND )));

//	qApp->processEvents();

	tabBar->getDesktop()->show();

	winManager->start();

#ifdef Q_OS_LINUX
	QProcess::startDetached("pnmixer", QStringList());
	QStringList arguments;
	arguments += "-i";
	arguments += "symbolic";
	QProcess::startDetached("cbatticon", arguments);
	arguments.clear();
	arguments += "-t";
	QProcess::startDetached("wicd-gtk", arguments);
	QProcess::startDetached("liberus-removablemanager", QStringList());
	//QProcess::startDetached("tuxum-config", QStringList());
#endif
	QTimer::singleShot( 100, this, SLOT(resize()));
	QTimer::singleShot( 1000, this, SLOT(setDockWindow()));
}

void Manager::setDockWindow()
{
	WinAction::setDockWindow( (HWND) tabBar->winId() );
	startWindowManager();
}

bool Manager::connectToServer(QString serverAddress, int serverPort)
{
	if ( ! LWClient::instance()->isConnected() )
	{
		if ( LWClient::instance()->connect( serverAddress, serverPort ) )
		{
			/* chroot environment now is managed by liberus-sesman
			res = LWClient::sendMessage( "mount_chroot?" +
					Settings::getValue( "ChrootHome" ) + "&" +
					Settings::getValue( "ChrootMount" ) + "&" +
					Settings::getUserName() + "&" +
					Settings::homePath() + "&" +
					Settings::getValue( "SambaServer" ) );
			Settings::logInfo( res );
			*/
			return true;
		} else {
			Settings::logError( tr("Cannot connect to lwserver") );
			return false;
		}
	} else {
		return true;
	}
}

void Manager::startWindowManager(bool disableAltTab)
{
#ifdef Q_OS_WIN
	Q_UNUSED( disableAltTab );
#endif
#ifdef Q_OS_LINUX
	QString command = "/usr/bin/liberuslauncher.sh";
	QString wm = Settings::getValue( "Advanced/PreferredWindowManager" );
	Settings::logInfo( QString( "Starting %1" ).arg( command ) );
	wmProcess = new QProcess();
	QStringList parameters;
	parameters += wm;
	if ( wm == "openbox" )
	{
		parameters += "--replace";
		parameters += "--config-file";
		if ( disableAltTab )
		{
			parameters += SHAREDIR "rc-no-alt-tab.xml";
		} else {
			parameters += SHAREDIR "rc.xml";
		}
	}
	if ( wm == "xfwm4" )
	{
		//parameters += "--daemon";
		parameters += "--replace";
		//parameters += "--sm-client-disable";
	}
	if ( wm == "sawfish" )
	{
		parameters += "--replace";
		parameters += "--no-rc";
	}
	wmProcess->start( command, parameters );
	if ( ! wmProcess->waitForStarted(NETWORK_START_WAIT_TIME) )
	{
		QMessageBox::warning(0, tr("Window Manager"), tr("Cannot start window manager"));
	} else {
		Settings::logInfo( QString( "%1 started succesfully" ).arg( wm ) );
	}

	if ( wmProcess->state() != QProcess::Running )
	{
		Settings::logInfo( QString( "%1 stopped unexpectedly with exit code %2" ).arg( wm ).arg(wmProcess->exitCode() ) );
	}
	if ( wmProcess->exitStatus() == QProcess::CrashExit )
	{
		QString wmError;
		switch (wmProcess->error() )
		{
			case QProcess::FailedToStart:
				wmError = "Failed to start process";
				break;
			case QProcess::Crashed:
				wmError = "Process crashed";
				break;
			case QProcess::Timedout:
				wmError = "Process timed out";
				break;
			case QProcess::WriteError:
				wmError = "Process write error";
				break;
			case QProcess::ReadError:
				wmError = "Process read error";
				break;
			case QProcess::UnknownError:
				wmError = "Failed to start process: unknown error";
				break;
			default:
				wmError = "No error";
		}
		Settings::logInfo( QString( "Error: %1" ).arg( wmError ) );
	}
	Settings::logInfo( QString( "Output: '%1' Errors: '%2'" ).arg( wmProcess->readAllStandardOutput().data() ).arg( wmProcess->readAllStandardError().data() ) );

#endif
}
void Manager::stopWindowManager()
{
#ifdef Q_OS_LINUX
	if ( wmProcess != 0 )
	{
		if ( wmProcess->state() == QProcess::Running )
		{
			wmProcess->kill();
			wmProcess->waitForFinished();
		}
		delete wmProcess;
		wmProcess = 0;
	}
#endif
}

void Manager::generateMimeAppsList()
{
#ifdef Q_OS_UNIX
	if ( QFile::exists( REMOTE_MIME_APPS_LIST ) )
	{
		QString mimeappsList = Settings::homePath() + QDir::separator() + QString( ".config/mimeapps.list" );
		if ( QFile::exists( mimeappsList ) )
		{
			QFile::rename( mimeappsList, mimeappsList + ".bak" );
		}

		QDir( Settings::homePath() ).mkdir ( ".config" );

		QSettings mimeapps( mimeappsList, QSettings::IniFormat );
		QFile remoteMimeAppsList( REMOTE_MIME_APPS_LIST );
		{
			if ( remoteMimeAppsList.open(QIODevice::ReadOnly) )
			{
				QTextStream in(&remoteMimeAppsList);
				while ( ! in.atEnd() )
				{
					QString line = in.readLine().simplified();
					if ( line.size() > 0 && ! line.startsWith( "#" ) )
					{
						mimeapps.setValue( "Added Associations/" + line, "liberus-remote.desktop" );
						mimeapps.setValue( "Default Applications/" + line, "liberus-remote.desktop" );
					}
				}
				remoteMimeAppsList.close();
			}
		}
		mimeapps.sync();
	}
#endif
}

Manager::~Manager()
{
	stopWindowManager();

	if ( appFileWatcher != nullptr)
	{
		delete appFileWatcher;
	}
}

QSize Manager::getDesktopSize()
{
	return QSize( QApplication::primaryScreen()->size().width(), QApplication::primaryScreen()->size().height() - TAB_BUTTON_HEIGHT);
}

void Manager::resize()
{
	qApp->processEvents();
#ifdef Q_OS_WIN
	WinInfo::hideTaskBar();
#endif
	tabBar->resize();
}

void Manager::clientConnected()
{
	resize();
}

void Manager::exitApp()
{
	tabBar->getDesktop()->saveGeometry();

	QString res;

	bool openApplications = false;

	if ( tabBar != nullptr)
	{
		if (tabBar->getButtonList()->count() > tabBar->menuButtonCount()) // There are open applications
		{
			openApplications = true;
		}

		if ( openApplications )
		{
			if ( QMessageBox::question(0, tr("Suspend session"), tr("There are currently running applications. Do you want to suspend session keeping them running?"), QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes)
			{
#ifdef Q_OS_WIN
				WTSDisconnectSession(WTS_CURRENT_SERVER_HANDLE, WTS_CURRENT_SESSION, FALSE);
#endif
#ifdef Q_OS_LINUX
				// TODO: Suspend session on linux
				Power::instance()->suspend();
#endif
			}
			return;
		}
	}

	winManager->stop();

	Application::deleteApplicationCache();

	if ( tabBar != nullptr )
	{
		//tabBar->close();
		delete tabBar;
	}

	if (LWClient::instance()->isConnected())
	{
		/* chroot environment now is managed by liberus-sesman
		if ( Settings::getBooleanValue( "ChrootEnvironment" ) )
		{
			res = LWClient::sendMessage( "umount_chroot?" + Settings::getUserName() );
			Settings::logInfo( res );
		}
		*/
		qDebug("Ending session...");
		res = LWClient::instance()->sendMessage( "end_session" );
		Settings::logInfo( res );
		LWClient::instance()->close();
	}
#ifdef Q_OS_WIN
	WinInfo::showTaskBar();
	ExitWindows(0, 0);
#endif
	qApp->exit(0);
	exit(0);
}

void Manager::timeout()
{
	launchingRemote = QString();
	QCursor * cursor = QApplication::overrideCursor();
	if ( cursor != nullptr )
	{
		if (cursor->shape() == Qt::WaitCursor)
		{
			QApplication::setOverrideCursor(Qt::ArrowCursor);
			tabBar->getDesktop()->setCursor(Qt::ArrowCursor);
		}
	}
	timer.stop();
	disconnect(&timer, SIGNAL(timeout()), this, SLOT(timeout()));
}

void Manager::createDesktopLinks(QString homeDir)
{
	QStringList folders;
	folders << "Downloads" << "Documents" << "Desktop";
	QDir home( homeDir );
	for ( int i = 0; i < folders.count(); i++ )
	{
		if ( ! QFile::exists(homeDir + QDir::separator() + folders[i]) )
		{
#ifdef Q_OS_LINUX
			home.mkdir( folders[i] );
#endif
			if ( folders[i] == "Desktop" )
			{
				QStringList links;
				links << "Downloads" << "Documents";
				for ( int i = 0; i < links.count(); i++ )
				{
#ifdef Q_OS_LINUX
					QFile::link( homeDir + QDir::separator() + links[i], homeDir + QDir::separator() + folders[i] + QDir::separator() + links[i]);
#endif
				}
			}
		}
	}
}

void Manager::launchApplication(Application *app)
{
	QCursor * cursor = QApplication::overrideCursor();
	if ( cursor != nullptr )
	{
		if (cursor->shape() == Qt::WaitCursor && timer.isActive() )
		{
			return;
		}
	}

	QApplication::setOverrideCursor(Qt::WaitCursor);
	tabBar->getDesktop()->setCursor(Qt::WaitCursor);
	QApplication::processEvents();

	timer.start(TIMEOUT * 1000);
	connect(&timer, SIGNAL(timeout()), this, SLOT(timeout()));

	app->launch();
	Settings::logInfo( QString("Launching application %1 with PID=%2")
			.arg(app->getName().toUtf8().data())
	                .arg((long long)app->getPid()));
}

void Manager::mapWindow(HWND id)
{
	WinAction::showWindow( id );
}

void Manager::mapWidget(QWidget * widget)
{
	widget->setGeometry(0, 0, getDesktopSize().width(), getDesktopSize().height());
	widget->show();
	WinAction::moveToFront( (HWND)widget->winId() );
}

void Manager::loadedFinished()
{
	timeout();
}

void Manager::openRemoteServerDir(QString targetDir)
{
	QString remoteServerAddress = Settings::getValue( "Server/LiberusRemoteAppServerAddress" );
	int remoteServerPort = Settings::getValue("Server/LiberusRemoteAppServerPort" ).toInt();
	if ( remoteServerPort > 0 && ! remoteServerAddress.isEmpty() )
	{
		LWClient * lwClient = new LWClient();
		QString token = Settings::getValue( TOKEN_FILE, "UserToken" );
		if ( lwClient->connect( remoteServerAddress, remoteServerPort ) )
		{
			if ( lwClient->authenticate( Settings::getFullUserName(), token ) )
			{
				QString remoteTargetDir = "\\\\" + Settings::getProfilesServerAddress() + "\\" + Settings::getFullUserName() + "\\TEMP\\" + targetDir + "\\";
				QString command = "_serverCommand_open_dir";
				lwClient->sendMessage( command, QStringList( remoteTargetDir ) );
			} else {
				qDebug("Cannot authenticate to remote application server with current user token");
			}
			LWSocket::closeSocket( lwClient->socket() );
		} else {
			qDebug("Cannot connect to remote application server %s:%d", remoteServerAddress.toUtf8().data(), remoteServerPort);
		}
		delete lwClient;
	}
}

void Manager::launchRemoteApplication( QString fileName, QString targetDir, QString remoteApplication )
{
	qDebug("Launch remote application %s %s %s", fileName.toUtf8().data(), targetDir.toUtf8().data(), remoteApplication.toUtf8().data());
	QString remoteServerAddress = Settings::getValue( "Server/LiberusRemoteAppServerAddress" );
	int remoteServerPort = Settings::getValue("Server/LiberusRemoteAppServerPort" ).toInt();
	if ( remoteServerPort > 0 && ! remoteServerAddress.isEmpty() )
	{
		LWClient * lwClient = new LWClient();
		QString token = Settings::getValue( TOKEN_FILE, "UserToken" );
		if ( lwClient->connect( remoteServerAddress, remoteServerPort ) )
		{
			if ( lwClient->authenticate( Settings::getFullUserName(), token ) )
			{
				QString remoteTargetDir = "\\\\" + Settings::getProfilesServerAddress() + "\\" + Settings::getFullUserName() + "\\TEMP\\" + targetDir;
				QStringList parameters;
				parameters += remoteTargetDir + "\\" + fileName;
				parameters += " ";
				QString command = "_serverCommand_open_file";
				lwClient->sendMessage( command, parameters );
			} else {
				qDebug("Cannot authenticate to remote application server with current user token");
			}
			lwClient->close();
		} else {
			qDebug("Cannot connect to remote application server %s:%d", remoteServerAddress.toUtf8().data(), remoteServerPort);
		}
		delete lwClient;
	}
}
