#include "settings.h"

#include <QDir>

static QHostAddress localAddress;
static QString userName;
static QString password;
static QString serverAddress;
static int serverPort = -1;
static QString profilesServerAddress;
static int profilesServerPort = -1;
static QMap<QString,QString> extensions;
static int windowManagerBehaviour = STANDARD;
static QString _homePath = QString("");
// logLevel: 0 = info, 9 = debug
static int logLevel = -1;

int Settings::getLogLevel()
{
	if ( logLevel == -1 )
	{
		bool ok;
		logLevel = getValue("Advanced/LogLevel").toInt(&ok);
		if ( ! ok )
		{
			logLevel = 0;
		}
	}
	return logLevel;
}

void Settings::init()
{
	QString behaviour = getValue("Desktop/WindowManagerBehaviour");

    if ( behaviour.toLower() == "fullscreen" )
    {
        windowManagerBehaviour = FULLSCREEN;

    } else {
        windowManagerBehaviour = STANDARD;
    }
}

int Settings::getWindowManagerBehaviour()
{
    return windowManagerBehaviour;
}

QHostAddress Settings::getLocalAddress()
{
	return localAddress;
}

QString Settings::getMimetype(QString extension)
{
    if ( ! extension.isEmpty() )
    {
		return extensions.value( extension ).simplified();
    } else {
        return QString();
    }
}

void Settings::setExtensions()
{
    QFile extensionsFile(EXTENSIONS_FILE);
    if (extensionsFile.open(QIODevice::ReadOnly))
    {
       QTextStream in(&extensionsFile);
       while (!in.atEnd())
       {
           QString res = in.readLine();
           extensions.insert( res.section(":", 0, 0), res.section(":", 1, 1));
       }
       extensionsFile.close();
    }
}

QStringList Settings::getApplications( QString mimetype )
{
    return getValue( MIMETYPES_FILE, mimetype ).split(":");
}

QStringList Settings::getFavoriteApplications()
{
	bool favoritesEnabled = getUserValue("FavoritesEnabled", QVariant(false) ).toBool();
	if ( ! favoritesEnabled )
	{
		QStringList favorites;
		favorites += "Google Chrome";
		favorites += "Chromium Web Browser";
		favorites += "Firefox ESR";
		favorites += "VLC media player";
		favorites += "Librework Client";
		favorites += "KCalc";
		favorites += "LibreOffice 5.3";
		setUserValue("Favorites", favorites);
	}
	return getUserValue("Favorites", QVariant()).toStringList();
}

void Settings::setFavoriteApplications(QStringList favoriteApps)
{
	setUserValue("FavoritesEnabled", true );
	setUserValue("Favorites", favoriteApps);
}

int Settings::getLocalOs()
{
	return RUNNING_OS;
}

QString Settings::getOsName(int os)
{
	switch (os)
	{
	case LINUX:
		return "Linux";
	case WINDOWS:
		return "Windows";
	default:
		return QObject::tr("Unknown");
	}
}

bool Settings::isLocalhost(QString address)
{
    if ( address.compare("127.0.0.1") )
    {
        return true;
    }
	if ( address.compare("localhost", Qt::CaseInsensitive ) )
    {
        return true;
    }
    if ( address.compare(QHostInfo::localHostName(), Qt::CaseInsensitive) )
    {
        return true;
    }
    if ( address.compare(QHostInfo::localHostName() + "." + QHostInfo::localDomainName(), Qt::CaseInsensitive) )
    {
        return true;
    }
    QList<QHostAddress> list = QNetworkInterface::allAddresses();
    for (int i = 0; i < list.count(); i++ )
    {
        if ( list[i].protocol() == QAbstractSocket::IPv4Protocol )
        {
            if ( QHostAddress(address) == list[i] )
            {
                return true;
            }
        }
    }
    return false;
}

void Settings::setLocalAddress(QHostAddress address)
{
	localAddress = address;
}

void Settings::setUserValue(QString key, QVariant value)
{
	QSettings settings(USER_CONFIG, QSettings::IniFormat);
	settings.setValue( key, value);
}

QVariant Settings::getUserValue(QString key, QVariant defaultValue)
{
	if ( key.isEmpty() )
	{
		return QVariant();
	}

	QSettings settings(USER_CONFIG, QSettings::IniFormat);
	QVariant value = settings.value( key, defaultValue );

	return value;
}

void Settings::setValue(QString file, QString key, QString value)
{
	QSettings settings(file, QSettings::IniFormat);
	settings.setValue( key, value);
	settings.sync();
}

QString Settings::getValue(QString primaryFile, QString secondaryFile, QString key)
{
    if ( key.isEmpty() )
    {
        return QString();
    }

	QString value;

	if ( QFile::exists(primaryFile) )
	{
		QSettings settings(primaryFile, QSettings::IniFormat);
		value = settings.value( key ).toString();
	}

	if (value.isEmpty())
	{
		QSettings settings(secondaryFile, QSettings::IniFormat);
		value = settings.value( key ).toString();
	}

	return value;
}

QString Settings::getApplicationValue( QString appName, QString key )
{
	return getValue( MIME_DIR + QString(QDir::separator()) + "applications", Settings::homePath()  + "/.liberus-desktop/mime/applications", appName + "/" + key );
}

QString Settings::getValue(QString key)
{
	return getValue(CONFIG_FILE, key);
}

QString Settings::getValue(QString file, QString key)
{
	if ( key.isEmpty() )
	{
		return QString();
	}

	QSettings settings(file, QSettings::IniFormat);
	QString value = settings.value( key ).toString();

	return value;
}

QStringList Settings::allKeys(QString file, QString group)
{
    QSettings settings(file, QSettings::IniFormat);
	if ( ! group.isEmpty() )
	{
		settings.beginGroup( group );
	}
	QStringList res = settings.allKeys();

	return res;
}

QString Settings::sharedPath()
{
	return getValue("SharedPath");
}

QString Settings::searchPath()
{
	return SEARCH_DIR;
}

QString Settings::windowsSharedPath()
{
	return "\\\\" + getValue("SambaServer") + getValue("SambaSharedPath");
}

QString Settings::windowsHomePath()
{
	return "\\\\" + getValue("SambaServer") + "\\" + getUserName();
}

void Settings::setHomePath( QString homePath )
{
	_homePath = homePath;
}

QString Settings::homePath()
{
	if ( ! _homePath.isEmpty() )
	{
		return _homePath;
	}
	QString hpath = getValue("Server/ProfilesServer");
	if (hpath.isEmpty())
	{
		_homePath = QDir::homePath();
		return _homePath;
	} else {
		hpath = QString(QDir::separator()) + QDir::separator() + hpath;
	}

	qDebug("getUserName(): '%s', userName: '%s'", getUserName().toUtf8().data(), userName.toUtf8().data());
	_homePath = hpath + QDir::separator() + getUserName();
	return _homePath;
}

int Settings::ldapServerPort()
{
	int value = getValue("LdapServerPort").toInt();

	if (value != 0)
		return value;
	else
		return 389;
}

bool Settings::getBooleanValue(QString key)
{
	QString value = getValue( key );
	bool ok;

	int number = value.toInt(&ok);

	if ( ok ) // is number
	{
		return ( number != 0 );
	} else {
		return ( value.toLower() == "true" );
	}
}

QStringList Settings::getApplicationsList()
{
#ifdef Q_OS_LINUX
	QStringList appList;
	QString apps = "/usr/share/applications";
	QDir appsDir( apps );
	QStringList filters;
	filters << "*.desktop";
	QStringList files = appsDir.entryList( filters );
	for ( int i = 0; i < files.count(); i++)
	{
		QSettings settings(apps + QDir::separator() + files[i], QSettings::IniFormat);
		QString name = settings.value("Desktop Entry/name").toString();
		if ( ! appList.contains( name ) )
		{
			appList += name;
		}
	}

	return appList;
#endif
#ifdef Q_OS_WIN
    QSettings settings(APPLICATIONS_FILE, QSettings::IniFormat);
	return settings.childGroups();
#endif
}

QStringList Settings::getIds(QString file)
{
    QSettings settings(file, QSettings::IniFormat);
	return settings.childGroups();
}

QStringList Settings::getValues(QString file, QString field)
{
    QStringList res;
    QStringList ids = getIds( file );
    for (int i = 0; i < ids.count(); i++)
    {
        res += getValue(file, ids[i] +"/" + field);
    }
    return res;
}

QString Settings::getUserName()
{
	if (userName.isEmpty())
	{
		QFile credentials( QDir::homePath()  + "/.credentials" );

		if (credentials.open(QIODevice::ReadOnly))
		{
			userName = credentials.readLine();
			userName.remove('\n');
		}
	}

	return userName;
}

QString Settings::getFullUserName()
{
	QString name = qgetenv("USER");
	if (name.isEmpty())
	{
		name = qgetenv("USERNAME");
	}

	if ( ! name.isEmpty() )
	{
		qDebug("Full username: %s", name.toUtf8().data());
		return name;
	} else {
		qDebug("Falling back to getUserName()");
		return getUserName();
	}
}

QString Settings::getPassword()
{
	if ( password.isEmpty() )
	{
		QFile credentials( Settings::homePath()  + "/.credentials" );
		if (credentials.open(QIODevice::ReadOnly))
		{
			credentials.readLine();
			password = credentials.readLine();
			password.remove('\n');
		}
	}

	if ( password.isEmpty() )
	{
		if (QFile::exists( TOKEN_FILE ))
		{
			password = Settings::getValue( TOKEN_FILE, "UserToken" );
		}
	}

	return password;
}

QString Settings::getServerAddress()
{
	qDebug("serverPort: %d", serverPort);
	if ( serverPort == -1 )
	{
		serverPort = getValue( "Server/LiberusServerPort" ).toInt();
		serverAddress = getValue( "Server/LiberusServerAddress" );
		qDebug("serverAddress: %s, serverPort: %d", serverAddress.toUtf8().data(), serverPort);
	}
	return serverAddress;
}

int Settings::getServerPort()
{
	qDebug("2serverPort: %d", serverPort);
	if ( serverPort == -1 )
	{
		QString serverAddress = getServerAddress();
		Q_UNUSED(serverAddress);
	}
	return serverPort;
}

QString Settings::getProfilesServerAddress()
{
	qDebug("getProfilesServerAddress... profilesServerPort=%d", profilesServerPort);
	if ( profilesServerPort == -1 )
	{
		qDebug("getServerPort: %d, getServerAddress: %s", getServerPort(), getServerAddress().toUtf8().data());
		if ( getServerPort() > 0 && ! getServerAddress().isEmpty() )
		{
			if ( LWClient::instance()->connect( getServerAddress(), getServerPort() ) )
			{
				qDebug("Authenticating with username: %s Token: %s", Settings::getFullUserName().toUtf8().data(), Settings::getValue( TOKEN_FILE, "UserToken" ).toUtf8().data());
				if ( LWClient::instance()->authenticate( Settings::getFullUserName(), Settings::getValue( TOKEN_FILE, "UserToken" ) ) )
				{
					QString completeProfilesServerAddress = LWClient::instance()->sendMessage("get_profiles_server").simplified();
					qDebug("CompleteProfilesServerAddress: %s", completeProfilesServerAddress.toUtf8().data());
					if ( completeProfilesServerAddress.isEmpty() )
					{
						profilesServerPort = 0;
					} else {
						profilesServerAddress = completeProfilesServerAddress.section(':', 0, 0 );
						QString profilesServerPortString = completeProfilesServerAddress.section(':', 1, 1);
						if ( profilesServerPortString.isEmpty() )
						{
							profilesServerPort = LIBERUS_DESKTOP_SERVER_PORT;
						} else {
							profilesServerPort = profilesServerPortString.toInt();
						}
					}
				} else {
					qDebug("Cannot authenticate with current user token");
				}
			} else {
				qDebug("Cannot connect to Liberus principal server");
			}
		} else {
			qDebug("There is no defined Liberus server in ini file");
		}
	}
	qDebug("Value: %s", profilesServerAddress.toUtf8().data());
	return profilesServerAddress;
}

int Settings::getProfilesServerPort()
{
	if ( profilesServerPort == -1 )
	{
		QString profilesServerAddress = getProfilesServerAddress();
		Q_UNUSED(profilesServerAddress);
	}
	return profilesServerPort;
}

QString Settings::getDesktopServer()
{
	QString res = getValue("Server/LiberusServerAddress");
	int port = getValue("Server/LiberusServerPort").toInt();
	if ( res.isEmpty() || port <= 0 )
	{
		return "localhost";
	} else {
		return res;
	}
}

QString Settings::getKioskApp()
{
	QString res = getValue("Advanced/KioskApp");
	return res;
}

void Settings::setUserName(QString user)
{
	userName = user;
}

void Settings::setPassword(QString pass)
{
	password = pass;
}

void Settings::setServerAddress(QString address)
{
	serverAddress = address;
}

void Settings::setServerPort(int port)
{
	serverPort = port;
}

void Settings::saveWidgetGeometry(QString name, QWidget * widget)
{
	setUserValue("WindowGeometry/" + name, widget->geometry());
}

void Settings::loadWidgetGeometry(QString name, QWidget * widget)
{
	QVariant geometry = getUserValue("WindowGeometry/" + name, QVariant());
	if ( ! geometry.isNull() )
	{
		widget->setGeometry( geometry.toRect() );
	}
}

void Settings::log(QString type, QString message)
{
	QString logString = QDateTime::currentDateTime().toString() + " - " + type + ": " + message;
	QFile logFile(Settings::homePath() + "/.liberus-desktop.log");
	if (logFile.open(QIODevice::WriteOnly | QIODevice::Append))
	{
		QTextStream out(&logFile);

		out << logString;
		out << "\n";
		logFile.close();
	}

	qDebug( logString.toUtf8().data() );
}

void Settings::logInfo(QString message)
{
	log("Info", message);
}

void Settings::logError(QString message)
{
	log("Error", message);
}

void Settings::logDebug(QString message)
{
	if ( getLogLevel() >= 9 )
	{
		log("Debug", message);
	}
}

QMap<QString, QString> Settings::desktopSettings(QString file)
{
	QMap<QString, QString> settings;
	QFile inFile( file );
	if ( inFile.open(QIODevice::ReadOnly) )
	{
		QTextStream in(&inFile);
		QString group;
		while (!in.atEnd())
		{
			QString line = in.readLine().simplified();
			if ( line.startsWith("[") && line.endsWith("]") )
			{
				group = line.mid(1, line.length() - 2) + "/";
			} else {
				int pos = line.indexOf('=');
				if ( pos > 0 )
				{
					settings[group + line.mid(0, pos)] = line.mid( pos + 1 );
				}
			}
		}
	}
	return settings;
}
