#include "systraymanager.h"

#ifdef Q_OS_LINUX
#include <X11/X.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <xcb/xcb.h>
#endif

SystrayManager::SystrayManager(HWND window, QObject *parent) : QObject( parent )
{
	systrayWinId = window;
	qApp->installNativeEventFilter( this );
#ifdef Q_OS_LINUX
	display = XOpenDisplay(NULL);
	init();
#endif
}

void SystrayManager::stop()
{
#ifdef Q_OS_LINUX
	XCloseDisplay(display);
#endif
}

bool SystrayManager::nativeEventFilter(const QByteArray & eventType, void * message, long * result)
{
#ifdef Q_OS_WIN
	Q_UNUSED( eventType );
	Q_UNUSED( message );
	Q_UNUSED( result );
#endif
#ifdef Q_OS_LINUX
	Q_UNUSED(result);

	if ( eventType == "xcb_generic_event_t" )
	{
		xcb_generic_event_t * xcbEvent = static_cast<xcb_generic_event_t *>(message);

		switch (xcbEvent->response_type & ~0x80)
		{
		case XCB_CLIENT_MESSAGE:
			return doClientMessage( message, result );
			break;

		case XCB_PROPERTY_NOTIFY:
			return propertyNotifyAction( message, result );
			break;
		case XCB_REPARENT_NOTIFY:
		case XCB_DESTROY_NOTIFY:
		case XCB_UNMAP_NOTIFY:
			break;
		default:
			//qDebug("Some event... Type: %d", xcbEvent->response_type);
			break;
		}
	}
#endif
	return false;
}

bool SystrayManager::propertyNotifyAction(void * message, long * result)
{
	Q_UNUSED( message );
	Q_UNUSED( result );
	//xcb_property_notify_event_t * e = static_cast<xcb_property_notify_event_t *>(message);
	//qDebug("Property notify %s on window: %d", XGetAtomName(display, e->atom), e->window );
	return false;
}

bool SystrayManager::doClientMessage(void * message, long * result)
{
#ifdef Q_OS_WIN
	Q_UNUSED( message );
	Q_UNUSED( result );
#endif
#ifdef Q_OS_LINUX
	xcb_client_message_event_t * e = static_cast<xcb_client_message_event_t *>(message);
	if ( e->type ==  XInternAtom(QX11Info::display(), "_NET_SYSTEM_TRAY_OPCODE", False)
		 && e->format == 32
		 && e->window == systrayWinId )
	{
		unsigned long opcode = ((unsigned long) e->data.data32[1]);
		Window window = (Window) ((unsigned long) e->data.data32[2]);
		qDebug("Detected _NET_SYSTEM_TRAY_OPCODE with opcode %d on window %d", (int)opcode, (int)window);
		switch ( opcode )
		{
		case 0: // SYSTEM_TRAY_REQUEST_DOCK
			qDebug("System tray request from window %d", (int)window);
			* result = 0;
			emit addIcon( (HWND) window );
			return true;
			break;
		case 1: // SYSTEM_TRAY_BEGIN_MESSAGE
			break;
		case 2: // SYSTEM_TRAY_CANCEL_MESSAGE
			break;
		default:
			break;
		}
	}
#endif
	return false;
}

void SystrayManager::init()
{
#ifdef Q_OS_LINUX
	QString name = QString("_NET_SYSTEM_TRAY_S%1").arg( DefaultScreen(display) );
	Atom systrayAtom = XInternAtom(display, name.toUtf8().data(), False);

	XSetSelectionOwner(display, systrayAtom, (Window)systrayWinId, CurrentTime);
	if (XGetSelectionOwner(display, systrayAtom) != (Window) systrayWinId )
	{
		qDebug("Cannot set selection owner to systray!");
	} else {
		XEvent m;
		m.type = ClientMessage;
		m.xclient.message_type = XInternAtom(display, "MANAGER", False);
		m.xclient.format = 32;
		m.xclient.data.l[0] = CurrentTime;
		m.xclient.data.l[1] = systrayAtom;
		m.xclient.data.l[2] = (Window)systrayWinId;
		m.xclient.data.l[3] = 0;
		m.xclient.data.l[4] = 0;
		XSendEvent(display, RootWindow(display, DefaultScreen(display)), False, StructureNotifyMask, &m);
	}
#endif
#ifdef Q_OS_WINDOWS
	WNDCLASSEXW class;
	static const WCHAR classname[] = {'S','h','e','l','l','_','T','r','a','y','W','n','d',0};

	icon_cx = GetSystemMetrics( SM_CXSMICON ) + 2*ICON_BORDER;
	icon_cy = GetSystemMetrics( SM_CYSMICON ) + 2*ICON_BORDER;

	/* register the systray listener window class */
	ZeroMemory(&class, sizeof(class));
	class.cbSize        = sizeof(class);
	class.style         = CS_DBLCLKS | CS_HREDRAW;
	class.lpfnWndProc   = tray_wndproc;
	class.hInstance     = NULL;
	class.hIcon         = LoadIconW(0, (LPCWSTR)IDI_WINLOGO);
	class.hCursor       = LoadCursorW(0, (LPCWSTR)IDC_ARROW);
	class.hbrBackground = (HBRUSH) COLOR_WINDOW;
	class.lpszClassName = classname;

	if (!RegisterClassExW(&class))
	{
		qDebug("Could not register SysTray window class");
		return;
	}

	tray_window = CreateWindowExW( WS_EX_NOACTIVATE, classname, NULL, WS_POPUP,
									0, GetSystemMetrics( SM_CYSCREEN ), 0, 0, 0, 0, 0, 0 );
	if (!tray_window)
	{
		qDebug("Could not create tray window\n");
		return;
	}
	SIZE size;
	NONCLIENTMETRICSW ncm;
	HFONT font;
	HDC hdc = GetDC( 0 );

	ncm.cbSize = sizeof(NONCLIENTMETRICSW);
	SystemParametersInfoW( SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICSW), &ncm, 0 );
	font = CreateFontIndirectW( &ncm.lfCaptionFont );
	/* FIXME: Implement BCM_GETIDEALSIZE and use that instead. */
	SelectObject( hdc, font );
	GetTextExtentPointA( hdc, "abcdefghijklmnopqrstuvwxyz", 26, &size );
	taskbar_button_width = size.cx;
	GetTextExtentPointW( hdc, start_label, lstrlenW(start_label), &size );
	/* add some margins (FIXME) */
	size.cx += 12 + GetSystemMetrics( SM_CXSMICON );
	size.cy += 4;
	ReleaseDC( 0, hdc );
	DeleteObject( font );

	tray_width = GetSystemMetrics( SM_CXSCREEN );
	tray_height = max( icon_cy, size.cy );
	start_button_width = size.cx;
	SetWindowPos( tray_window, HWND_TOPMOST, 0, GetSystemMetrics( SM_CYSCREEN ) - tray_height,
				tray_width, tray_height, SWP_NOACTIVATE | SWP_SHOWWINDOW );
#endif
}
