#include "manager.h"
#include "threadbutton.h"
#include <QDir>
#include "wininfo.h"

#ifdef Q_OS_LINUX
ThreadButton::ThreadButton (QWidget * widget, int type, QPixmap pixmap, QString label, QString text, HWND window, Q_PID pid, bool isRemote ) : QPushButton()
#endif
#ifdef Q_OS_WIN
ThreadButton::ThreadButton (QWidget * widget, int type, QPixmap pixmap, QString label, QString text, HWND window, unsigned long pid, bool isRemote ) : QPushButton()
#endif
{
	setWindowFlags( this->windowFlags() | Qt::BypassWindowManagerHint | Qt::WindowDoesNotAcceptFocus | Qt::WindowStaysOnBottomHint );
	setFocusPolicy(Qt::NoFocus);
	connect(this, SIGNAL(clicked()), this, SLOT(buttonClicked()));
	setFixedSize( QSize(TAB_BUTTON_WIDTH, TAB_BUTTON_HEIGHT) );
	setIconSize( QSize(TAB_BUTTON_WIDTH - ICONMARGIN, TAB_BUTTON_WIDTH - ICONMARGIN) );
	this->widget = widget;
	this->type = type;
	this->pixmap = pixmap;
	this->label = label;
	this->text = text;
	this->window = window;
	this->pid = pid;
	this->remote = isRemote;

	tip = new ToolTip();

	setAccessibleName( label );

	selected = false;

	rightClickWidget = nullptr;
	leftClickWidget = nullptr;
	index = -1;
}

ThreadButton::~ThreadButton()
{
	delete tip;
}

void ThreadButton::setSelected( bool selected )
{
	this->selected = selected;
#ifdef Q_OS_WIN
    //forceRepaint();
#else
	repaint();
#endif
}

void ThreadButton::forceRepaint()
{
	QSize _size = size();
	setFixedSize(size().width() - 1, size().height() - 1);
	setFixedSize(_size);
}

QSize ThreadButton::buttonSize()
{
	return QSize( TAB_BUTTON_WIDTH, TAB_BUTTON_HEIGHT );
}

void ThreadButton::buttonClicked()
{
	if ( leftClickWidget != nullptr )
	{
		emit threadButtonClicked(leftClickWidget);
	} else  if ( widget != nullptr ) {
		emit threadButtonClicked(widget);
	} else {
		emit threadButtonClicked(window);
	}
}

void ThreadButton::closeButton()
{
	emit closeThreadButton( index );
}

void ThreadButton::closeApplication()
{
	tip->hide();
	WinAction::closeWindow( window );
}

void ThreadButton::terminateApplication()
{
	tip->hide();

	if ( QMessageBox::critical(nullptr, tr("Terminate application"),
				tr("The application is to be terminated. All pending data will be lost. Are you sure you want to continue?"),
				QMessageBox::Yes | QMessageBox::No, QMessageBox::No)  == QMessageBox::Yes )
	{
		WinAction::terminateWindow( window );
	}
}

void ThreadButton::setPixmap(QPixmap pixmap)
{
	this->pixmap = pixmap;
	this->update();
}

void ThreadButton::mousePressEvent( QMouseEvent *e )
{
	if( type == LOGOMENUBUTTON )
	{
			QSize widgetSize = widget->size();
			QSize desktopSize = Manager::getDesktopSize();
			QRect widgetGeometry( QPoint(5, desktopSize.height() - widgetSize.height() - 5) , widgetSize);
			widget->setGeometry( widgetGeometry );
			widget->show();
			widget->setFocus();
	}

	if ( type == THREADBUTTON )
	{
		if ( e->button() == Qt::RightButton )
		{
			QMenu * rightClickMenu = new QMenu( this );
			QMenu * rightClickSubMenu = new QMenu( this );
			rightClickMenu->addAction( QIcon(":/icons/32x32/cancel.png"), tr("&Close"), this, SLOT(closeApplication()));
			rightClickMenu->addSeparator();
			rightClickMenu->addAction( tr("&More options"), rightClickSubMenu, SLOT(show()));
			connect( rightClickMenu, SIGNAL(aboutToHide()), rightClickMenu, SLOT(deleteLater()));

			rightClickMenu->adjustSize();
			QPoint point = mapToGlobal(QPoint(0, 0));
			point.setY( point.y() - rightClickMenu->height() );
			rightClickMenu->move( point );

			rightClickSubMenu->addAction( QIcon(":/icons/32x32/terminate.png"), tr("&Force termination"), this, SLOT(terminateApplication()));
			rightClickSubMenu->adjustSize();
			QPoint subPoint = mapToGlobal(QPoint(0, 0));
			subPoint.setY( subPoint.y() - rightClickSubMenu->height() );
			rightClickSubMenu->move( subPoint );

			rightClickMenu->show();
		}
	}
	QPushButton::mousePressEvent( e );
}

void ThreadButton::mouseDoubleClickEvent(QMouseEvent *e)
{
	QPushButton::mouseDoubleClickEvent( e );
	emit threadButtonDoubleClicked();
}

void ThreadButton::paintEvent ( QPaintEvent * e)
{
	Q_UNUSED( e );

	QPainter p(this);

	p.setRenderHints( QPainter::SmoothPixmapTransform, true );
	p.setRenderHints( QPainter::Antialiasing, true );

	QPixmap pPixmap = pixmap;
	/*QRadialGradient gradient(50, 50, 50, 50, 50);
	gradient.setColorAt(0, Qt::gray);
	gradient.setColorAt(1, Qt::darkGray);
	QBrush brush( gradient );*/

	if (selected && type != EXITBUTTON)
	{
		//QPushButton::paintEvent( e );
		//p.setBackgroundMode(Qt::OpaqueMode);
		//p.setBackground( brush );
	}

	if ( type == EXITBUTTON )
	{
		if ( isDown() )
		{
			pPixmap =  QPixmap(ICON_DIR_24 + QString(QDir::separator()) + "power-pressed.png");
		}
	}

	if ( ! pPixmap.isNull() )
	{
		int pressedMargin = 0;

		if ( isDown() && type != EXITBUTTON )
		{
			pressedMargin = BUTTON_PRESSED_MARGIN;
		}

		int wMargin = ( TAB_BUTTON_WIDTH - MENU_ICON_SIZE ) / 2;
		int hMargin = ( TAB_BUTTON_HEIGHT - MENU_ICON_SIZE ) / 2;

		if ( type == LOGOMENUBUTTON )
		{
			wMargin = ( TAB_BUTTON_WIDTH - LOGO_MENU_ICON_SIZE ) / 2;
			hMargin = ( TAB_BUTTON_HEIGHT - LOGO_MENU_ICON_SIZE ) / 2;
		}

		QPoint upLeft( wMargin + pressedMargin, hMargin + pressedMargin);
		QPoint downRight( TAB_BUTTON_WIDTH - pressedMargin - wMargin - 1, TAB_BUTTON_HEIGHT - pressedMargin - hMargin - 1 );
		QRect rectangle (upLeft, downRight);
		//qDebug("Pixmap width: %d, height: %d, rectangle width: %d, height: %d for %s - %s", pPixmap.width(), pPixmap.height(), rectangle.width(), rectangle.height(), label.toUtf8().data(), text.toUtf8().data());
		if (selected && type != EXITBUTTON)
		{
			QBrush brush(Qt::SolidPattern);
			brush.setColor(QColor(TABSELECTIONCOLOR));
			p.fillRect( QRect( QPoint(0,0), size() ), brush );
		}
		p.drawPixmap( rectangle, pPixmap );
	}
}

bool ThreadButton::event(QEvent *event)
{
	if ( event->type() == QEvent::Enter )
	{
		switch ( type )
		{
		case EXITBUTTON:
			tip->setText( tr("Exit Liberus") );
			break;
		case MENUBUTTON:
		case DESKTOPBUTTON:
		case LOGOMENUBUTTON:
			tip->setText( "" );
			break;
		default:
			tip->setText( WinInfo::getTitle( window ) );
			break;
		}
		if ( tip->text().isEmpty() )
		{
			tip->hide();
			return QPushButton::event( event );
		}

		tip->adjustSize();
		int tipWidth = tip->width();
		int tipHeight = tip->height();
		int posX = mapToGlobal(QPoint(0, 0)).x();
		int posY = mapToGlobal(QPoint(0, 0)).y();
		int x = posX + ( this->width() / 2 ) - (tipWidth / 2);
		int y = posY - tipHeight - 7;
		if ( x < 1 ) x = 1;
		if ( x + tipWidth >= qApp->primaryScreen()->geometry().width() ) x = qApp->primaryScreen()->geometry().width() - tipWidth - 1;
		tip->move(x, y);
		tip->show();
	}

	if ( event->type() == QEvent::Leave )
	{
		tip->hide();
	}

	return QPushButton::event( event );
}
