#include "touchconfig.h"

TouchConfig::TouchConfig(QWidget *parent ) : QDialog( parent )
{
	background = nullptr;
	QStringList arguments;
	arguments += "-n"; // non-interactive sudo
	arguments += "liberus-touchscreen.sh";
	QProcess::startDetached( "sudo", arguments );
	setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
	init ();
}

bool TouchConfig::isAvailable()
{
	return QFile::exists( "/usr/bin/mrpdi-cli" );
}

void TouchConfig::init()
{
	layout = new QGridLayout(this);
	layout->setSpacing(10);
	layout->setMargin(20);

	QFont bFont;
	bFont.setFamily( Settings::getUserValue("BoldFont", LIBERUS_FONT_BOLD).toString() );

	QLabel * devicesLabel = new QLabel(tr("Device:"), this);
	devicesLabel->setMinimumWidth( 150 );
	devicesLabel->setFont( bFont );
	devices = new QComboBox( this );
	devices->setMinimumWidth( 250 );

	calibrate = new QPushButton( tr("Calibrate"), this);
	calibrate->setFlat( true );
	calibrate->setStyleSheet(STANDARDBUTTONSTYLE);
	connect( calibrate, SIGNAL( clicked() ), this, SLOT( calibrateBackgroundSlot() ) );

	refresh = new QPushButton( tr("Refresh"), this);
	refresh->setStyleSheet(STANDARDBUTTONSTYLE);
	refresh->setMinimumSize( 100, 24 );
	connect( refresh, SIGNAL( clicked() ), this, SLOT( loadSettings() ) );

	layout->addWidget(devicesLabel, 1, 0);
	layout->addWidget(devices, 1, 1, 1, 5);
	layout->addWidget(calibrate, 1, 6, 1, 1);
	layout->addWidget(refresh, 5, 5);

	for (int i = 0; i < 8; i++)
	{
		layout->setRowStretch(i, 1);
		layout->setColumnStretch(i, 1);
	}

	loadSettings();
}

void TouchConfig::calibrateBackgroundSlot()
{
	if ( ! devices->currentText().isEmpty() )
	{
		background = new QLabel( tr("Calibrating touch screen. Wait..."), nullptr );
		QFont bFont;
		bFont.setFamily( Settings::getUserValue("BoldFont", LIBERUS_FONT_BOLD).toString() );
		background->setFont( bFont );
		background->setAlignment( Qt::AlignCenter );
		background->showFullScreen();
		QApplication::processEvents();
		QTimer::singleShot( 100, this, SLOT( calibrateSlot() ) );
	}
}

void TouchConfig::calibrateSlot()
{
	QString calibrationFile;
	// Check calibration utility needed
	if ( devices->currentData( DEVICETYPE ).toString() == "iqboard" )
	{
		calibrationFile = "/usr/share/BoardDriver/DVT/Config/config.xml";
	}

	if ( devices->currentData( DEVICETYPE ).toString() == "mrpdi" )
	{
		calibrationFile = "/etc/mrpdi/settings.conf";
	}

	if ( ! calibrationFile.isEmpty() )
	{
		// Remove calibration file if exists
		if ( QFile::exists( calibrationFile ) )
		{
			QStringList arguments;
			arguments += getMacAddress();
			arguments += calibrationFile;
			QString ret = LWClient::instance()->sendMessage( "remove_pxe_config_file", arguments );
			if ( ret.startsWith("Error") )
			{
				QString errorMessage = ret;
				QMessageBox::critical( nullptr, tr("Error setting up calibration"), errorMessage, QMessageBox::Ok );
				return;
			}
		}
	}

	calibrateProcess = new QProcess;
	QStringList arguments;
	arguments += "-n"; // non-interactive sudo
	arguments += "liberus-touchcalibrate.sh";
	arguments += devices->currentData().toString();
	connect( calibrateProcess, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(calibrateFinished()) );
	calibrateProcess->start( "sudo", arguments );
	if ( ! calibrateProcess->waitForStarted( TIMEOUT * 1000 ) )
	{
		if ( background != nullptr )
		{
			delete background;
			background = nullptr;
		}
	}
}

void TouchConfig::calibrateFinished()
{
	int ret = calibrateProcess->exitCode();

	if ( background != nullptr )
	{
		delete background;
		background = nullptr;
	}
	delete calibrateProcess;

	QApplication::processEvents();

	QString errorMessage;
	if ( ret != 0 )
	{

		switch ( ret )
		{
		case 1:
			errorMessage = tr("Cannot remove calibration file");
			break;
		case 2:
			errorMessage = tr("Calibration file not generated");
			break;
		case 3:
			errorMessage = tr("Calibration process failed");
			break;
		default:
			errorMessage = tr("Unknown error calibrating touch screen");
		}

	} else {
		QString macAddress = getMacAddress();

		if ( macAddress.isEmpty() )
		{
			errorMessage = tr("Cannot read MAC address on this computer");
		}
		QString calibrationFile;
		if ( devices->currentData( DEVICETYPE ).toString() == "iqboard" )
		{
			calibrationFile = "/usr/share/BoardDriver/DVT/Config/config.xml";
		}

		if ( devices->currentData( DEVICETYPE ).toString() == "mrpdi" )
		{
			calibrationFile = "/etc/mrpdi/settings.conf";
		}
		QFile settingsFile( calibrationFile );
		if ( errorMessage.isEmpty() && settingsFile.open( QIODevice::ReadOnly ) )
		{
			QStringList arguments;
			arguments += macAddress;
			arguments += calibrationFile;
			arguments += settingsFile.readAll();
			QString ret = LWClient::instance()->sendMessage( "set_pxe_config_file", arguments );
			if ( ret.startsWith("Error") )
			{
				errorMessage = ret;
			}
			settingsFile.close();
		} else {
			errorMessage = tr("Cannot open calibration file");
		}
	}

	if ( ! errorMessage.isEmpty() )
	{
		QMessageBox::critical( nullptr, tr("Error calibrating touch screen"), errorMessage, QMessageBox::Ok );
	} else {
		QMessageBox::information( nullptr, tr("Calibration successful"), tr("Touch screen successfuly calibrated"), QMessageBox::Ok );
	}
}

QString TouchConfig::getMacAddress()
{
	QString macAddress;
	foreach(QNetworkInterface netInterface, QNetworkInterface::allInterfaces())
	{
		if ( ! ( netInterface.flags() & QNetworkInterface::IsLoopBack) )
		{
			macAddress = netInterface.hardwareAddress();
		}
	}
	return macAddress;
}

void TouchConfig::loadSettings()
{
	devices->clear();

	// mrpdi drivers
	QProcess devicesProcess;
	devicesProcess.start( "mrpdi-cli", QStringList("--list") );
	devicesProcess.waitForStarted();
	devicesProcess.waitForReadyRead();
	devicesProcess.waitForFinished();
	QList<QByteArray> devicesList = devicesProcess.readAll().split('\n');

	for ( int i = 0; i < devicesList.count(); i++ )
	{
		int posDev = devicesList[i].indexOf( "dev:" );
		int posAddress = devicesList[i].indexOf( "address:" );
		int posId = devicesList[i].indexOf( "id:" );

		if ( posDev >= 0 && posAddress >= 0  && posId >= 0 )
		{
			QString deviceString = devicesList[i].mid( posDev + 5, posAddress - posDev - 5 ).simplified();
			QString deviceAddress = devicesList[i].mid( posAddress + 9, posId - posAddress - 9 ).simplified();
			devices->addItem( deviceString, deviceAddress );
			devices->setItemData( devices->currentIndex(), "mrpdi", DEVICETYPE );
		}
	}

	// IQBoard drivers
	if ( checkUsb( "10c4", "ea60") ) // IQBoard
	{
		devices->addItem( "IQBoard", "ea60" );
		devices->setItemData( devices->currentIndex(), "iqboard", DEVICETYPE );
	}

	if ( devicesList.count() > 0 )
	{
		devices->setCurrentIndex( 0 );
		calibrate->setEnabled( true );
	} else {
		calibrate->setEnabled( false );
	}
}

bool TouchConfig::checkUsb(QString vendor, QString product)
{
	return checkUsb( vendor, QStringList( product ) );
}

bool TouchConfig::checkUsb(QString vendor, QStringList products)
{
	bool ret = false;
	for ( int i = 0; i < products.count(); i++ )
	{
		QProcess lsUsbProcess;
		QStringList arguments;
		arguments += "-d";
		arguments += vendor + ":" + products[i];
		lsUsbProcess.start("lsusb", arguments);
		lsUsbProcess.waitForFinished();
		ret = ( ret || lsUsbProcess.exitCode() == 0 );
	}
	return ret;
}
